# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

GIRAFFE_NOT_FOUND = False
try:
    from . import _encoder
except ImportError:
    GIRAFFE_NOT_FOUND = True

import collections

from .config import *
from .constants import *
from .errors import *
from .logging import *
from .utils import *


__all__ = ['Connection']


class Connection(object):
    def __init__(self, host=None, username=None, password=None, log_level=INFO, config=None,
            key_file=None, dsn=None, protect=False, mload_session=False):
        if GIRAFFE_NOT_FOUND:
            raise GiraffeNotFound("giraffez module was not compiled with package")
        #: Log level initially set to SILENCE to ensure that using the
        #: Python API does not produce log output unless expressly set
        log.level = log_level

        self.config = config
        self.dsn = dsn
        self.protect = protect

        self.host = host

        #: Stores options for log output
        self.options = ConnectionOptions()

        #: Attribute identifies mload sessions to suppress duplicate log
        #: output since mload requires two Connection objects
        self.mload_session = mload_session

        if host is None or username is None or password is None:
            if host or username or password:
                show_warning(("You must set all arguments to overide default (host, username, "
                    " password)"), UserWarning)
            with Config(config, key_file=key_file) as c:
                conn = c.get_connection(dsn)
            self.dsn = conn.get("name")
            host = conn.get("host", None)
            if host is None:
                raise suppress_context(ConfigurationError("Connection '{}' missing host value".format(self.dsn)))
            username = conn.get("username")
            if username is None:
                raise suppress_context(ConfigurationError("Connection '{}' missing username value".format(self.dsn)))
            password = conn.get("password")
            if password is None:
                raise suppress_context(ConfigurationError("Connection '{}' missing password value".format(self.dsn)))
        try:
            if not self.mload_session:
                log.info("Connection", "Connecting to data source '{}' ...".format(self.dsn))
            self._connect(host, username, password)
            if not self.mload_session:
                log.info("Connection", "Connection to '{}' established successfully.".format(self.dsn))
        except InvalidCredentialsError as error:
            if self.protect:
                Config.lock_connection(self.config, self.dsn)
            raise error

    def _connect(self, host, username, password):
        pass

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        if not self.mload_session:
            log.info("Connection", "Closing Teradata connection ...")
        self.close()
        if not self.mload_session:
            log.info("Connection", "Connection to '{}' closed.".format(self.dsn))


class ConnectionOptions(object):
    def __init__(self):
        self._options = collections.OrderedDict()

    @property
    def width(self):
        if not self._options.keys():
            return 0
        return len(max(self._options, key=lambda x: len(x)))

    def clear(self):
        self._options = {}

    def set(self, name, value, priority=1):
        self._options[name] = (priority, value)

    def unset(self, name, value):
        return self._options.pop(name, None)

    def __call__(self, name, value, priority=1):
        return self.set(name, value, priority)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        option = '  {}{} => "{}"'
        r = []
        for k, (priority, v) in sorted(self._options.items(), key=lambda x: str(x[1])):
            offset_width = self.width - len(k)
            value = str(v).replace('\n', '\n    ')
            r.append(option.format(k, " "*offset_width, value))
        return "\n".join(r)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import cmd as _cmd
import os
import sys
import time
import atexit
import itertools
import functools
import threading

from .cmd import *
from .constants import *
from .errors import *
from .fmt import *
from .logging import *
from .parser import *
from .utils import *


__all__ = ['GiraffeShell']


setup_logging()


class Spinner(threading.Thread):
    def __init__(self):
        super(Spinner, self).__init__()
        self._stop_event = threading.Event()
        self._ended = threading.Event()
        self._lock = threading.Lock()

    def run(self):
        spinner = itertools.cycle(["-", "/", "|", "\\"])
        while True:
            sys.stderr.write(next(spinner))
            time.sleep(0.1)
            if self.stopped:
                sys.stderr.flush()
                sys.stderr.write("\b")
                break
            sys.stderr.flush()
            sys.stderr.write("\b")
        self._ended.set()

    @property
    def running(self):
        return self._ended.isSet()

    def stop(self):
        self._stop_event.set()

    @property
    def stopped(self):
        return self._stop_event.isSet()


def spinner(f):
    @functools.wraps(f)
    def wrapper(*args, **kwds):
        spinner = Spinner()
        spinner.start()
        try:
            result = f(*args, **kwds)
            spinner.stop()
            spinner.join()
        except Exception as error:
            spinner.stop()
            spinner.join()
            raise error
        return result
    return wrapper

def save_history():
    import os
    import readline
    path = os.path.expanduser("~/.giraffehistory")
    readline.write_history_file(path)


class GiraffeShell(_cmd.Cmd):
    """Interactive shell for giraffez"""

    prompt = "giraffez> "
    intro = "giraffez-shell version: {}".format(get_version_info()[1])
    table_output = True

    start_chars = ("(", "{")
    end_chars = (")", "}")

    help_commands = ("help session", "help table", "help column", "help view")

    def __init__(self, cmd):
        import readline
        _cmd.Cmd.__init__(self)
        path = os.path.expanduser("~/.giraffehistory")
        if os.path.exists(path):
            readline.read_history_file(path)
        atexit.register(save_history)
        self.cmd = cmd

    @timer
    @spinner
    def command(self, query):
        return self.cmd.execute_one(query)

    def complete_table(self, text, line, begidx, endidx):
        table_values = ["on", "off"]
        completions = [f for f in table_values if f.startswith(text)]
        return completions

    def default(self, line):
        if line.startswith(".help"):
            line = line[1:]
        try:
            sc = sum([line.count(c) for c in self.start_chars])
            ec = sum([line.count(c) for c in self.end_chars])
            if sc != ec:
                while True:
                    q = input("       > ")
                    if not q:
                        break
                    line += q
                    sc = sum([line.count(c) for c in self.start_chars])
                    ec = sum([line.count(c) for c in self.end_chars])
                    if sc == ec:
                        break
        except KeyboardInterrupt:
            sys.stderr.write("\n")
            return
        try:
            result = self.command(line)
        except TeradataError as error:
            log.write("{0!r}".format(error))
            return
        if not result:
            log.write("No results")
            return
        if self.table_output and not (line.lower().startswith("show") or line.lower().startswith("help")):
            log.write()
            result.rows.insert(0, result.columns.names)
            sys.stdout.write(format_table(result.rows))
            sys.stdout.write("\n")
        else:
            for row in result:
                log.write("\t".join([str(x) for x in row]))

    def do_EOF(self, line):
        return True

    def do_exit(self, line):
        sys.exit()

    def do_horse(self, line):
        """It totally prints a horse."""
        log.write(ascii_giraffe)

    def do_table(self, line):
        """Display results in table format"""
        if len(line) > 0:
            if line.strip().lower() == "on":
                log.write("Table ON")
                self.table_output = True
                return
            elif line.strip().lower() == "off":
                log.write("Table OFF")
                self.table_output = False
                return
        log.write("Table output: {}".format("ON" if self.table_output else "OFF"))

    def precmd(self, line):
        if line.strip().lower().startswith(self.help_commands):
            line = ".{}".format(line)
        return _cmd.Cmd.precmd(self, line)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import base64

import yaml

from .encrypt import *
from .errors import *
from .io import *
from .logging import *
from .parser import *
from .utils import *

from ._compat import *


__all__ = ['default_config', 'Config', 'message_write_default']


message_write_default = """Successfully created file '{0}'
Set the username and password using the config module:

    giraffez config --set connections.db1.host <host>
    giraffez config --set connections.db1.username <username>
    giraffez config --set connections.db1.password <password>"""

default_config = {
    "connections": {
        "default": "db1",
        "db1": {
            "host": None,
            "username": None,
            "password": None
        }
    },
    "colorful": True
}

def _get_config_file(conf=None):
    if conf is None:
        conf = home_file(".girafferc")
    else:
        conf = os.path.abspath(conf)
    if not file_exists(conf):
        raise ConfigNotFound(("Config file '{}' does not exist and credentials were not "
            "otherwise specified.").format(conf))
    return conf

def _get_key_file(key_file=None):
    if key_file is None:
        key_file = home_file(".giraffepg")
    else:
        key_file = os.path.abspath(key_file)
    if not file_exists(key_file):
        raise KeyNotFound("Key file is required to use giraffez config.")
    return key_file

def _check_file_permissions(path, perms):
    if sys.platform == 'win32':
        return True
    if file_permissions(path) != perms:
        perms = oct(perms).replace('o', '')
        command = colors.fail("chmod {} {}".format(perms, path))
        raise ConfigurationError(("File '{}' not set with secure file-level "
            "permissions. Try running chmod to correct this issue:\n\t{}").format(
            path, command))
    return True

def _read_config_dict(d, fn, path=[], level=0):
    keywords = ["pass", "password", "apikey", "key"]
    level += 1
    for k, v in d.items():
        path.append(k)
        if isinstance(v, dict):
            _read_config_dict(v, fn, path, level)
        else:
            if k in keywords:
                if v and v.startswith("ENCRYPTED:"):
                    v = v.rsplit(":", 1)[1]
                    d[k] = str(fn(v))
            if path:
                if path[0] == "secure" and v.startswith("ENCRYPTED:"):
                    v = v.rsplit(":", 1)[1]
                    d[k] = str(fn(v))
        path.remove(k)


class Config(object):
    """
    An object for reading, writing, and encrypting giraffez configuration
    files (YAML files). Uses an encryption key (required) to encrypt and
    decrypt values keyed as 'password', 'pass', 'apikey', or 'key', as
    well as any values nested under the key 'secure' (e.g. setting
    'secure.teradata.user' will encrypt the value.)

    Implements the :code:`__enter__` and :code:`__exit__` magic methods
    so that it can be used by Python's :code:`with` context-handler:

    .. code-block:: python

       with Config() as conf:
           conn = conf.get_connection() # get default connection
       username = conn.get('username')
       password = conn.get('password')

    :param str conf: A path to the configuration file to open.
        Defaults to ~/.girafferc
    :param str mode: Defaults to 'r' for read-only. If not changed,
        configuration will not be writeable while open.
    :param str key_file: A path to the key file to open.
        Defaults to ~/.giraffepg

    :raises `giraffez.errors.KeyNotFound`: if default key_file is not
        found and a valid one is not provided.
    """

    def __init__(self, conf=None, mode="r", key_file=None):
        if conf is None:
            conf = home_file(".girafferc")
        self._config_file = _get_config_file(conf)
        if key_file is None:
            self._key_file = home_file(".giraffepg")
        self._key_file = _get_key_file(key_file)
        self.mode = mode
        _check_file_permissions(self._config_file, 0o600)
        _check_file_permissions(self._key_file, 0o400)
        self.settings = self.read()
        self.cipher = Crypto.from_key_file(self._key_file)

    @property
    def connections(self):
        """
        Return a :code:`dict` of connections from the configuration settings.

        :raises `giraffez.errors.ConfigurationError`: if connections are not present
        """
        if "connections" not in self.settings:
            raise ConfigurationError("Could not retrieve connections from config file '{}'.".format(self._config_file))
        return self.settings.get("connections")

    def decrypt(self, value, path=[]):
        if isinstance(value, dict):
            _read_config_dict(value, self.cipher.decrypt, path)
        elif isinstance(value, basestring):
            if value.startswith("ENCRYPTED:"):
                value = value.rsplit(":", 1)[1]
                value = self.cipher.decrypt(value)
        return value

    def encrypt(self, value, path):
        if path:
            key = path[-1].lower()
            if key in ["pass", "password", "apikey", "key"] or path[0] in ["secure"]:
                if not isinstance(value, basestring):
                    raise ConfigurationError("Cannot encrypt a value of type '{}'.".format(type(value)))
                value = "ENCRYPTED:{}".format(ensure_str(self.cipher.encrypt(value)))
        return value

    def get_connection(self, dsn=None):
        """
        Retrieve a connection by the given :code:`dsn`, or the default connection.

        :param str dsn: The name of the connection to retrieve. Defaults to :code:`None`,
            which retrieves the default connection.
        :return: A dict of connection settings
        :raises `giraffez.errors.ConfigurationError`: if :code:`dsn` does not exist and a
        :raises `giraffez.errors.ConnectionLock`: if the corresponding connection is currently locked
            default is not set.
        """
        if dsn is None:
            dsn = self.connections.get("default", None)
            if dsn is None:
                raise ConfigurationError("No default DSN set")
        connection = self.connections.get(dsn, None)
        if connection is None:
            raise ConfigurationError("DSN '{}' does not exist".format(dsn))
        connection = self.decrypt(connection.copy())
        if connection.get('lock', 0) > 1:
            raise ConnectionLock(dsn)
        connection['name'] = dsn
        return connection

    def get_value(self, key, default={}, nested=True, decrypt=True):
        """
        Retrieve a value from the configuration based on its key. The key
        may be nested.

        :param str key: A path to the value, with nested levels joined by '.'
        :param default: Value to return if the key does not exist (defaults to :code:`dict()`)
        :param bool decrypt: If :code:`True`, decrypt an encrypted value before returning
            (if encrypted).  Defaults to :code:`True`.
        """
        key = key.lstrip()
        if key.endswith("."):
            key = key[:-1]
        if nested:
            path = key.split(".")
            curr = self.settings
            for p in path[:-1]:
                curr = curr.get(p, {})
            try:
                value = curr[path[-1]]
            except KeyError:
                return default
            value = self.decrypt(value, path)
            return value
        else:
            return self.settings.get(key, default)

    def list_value(self, decrypt=False):
        """
        Return the contents of the configuration as a :code:`dict`. Depending on
        the structure of the YAML settings, the return value may contain nested
        :code:`dict` objects.

        :param bool decrypt: If :code:`True`, decrypt the contents before returning.
        :return: (potentially) nested :code:`dict` of keys and values, as parsed from
            the configuration file YAML contents.
        """
        if decrypt:
            settings = self.decrypt(self.settings)
        else:
            settings = self.settings
        return yaml.dump(settings, default_flow_style=False)

    @classmethod
    def lock_connection(cls, conf, dsn, key=None):
        """
        A class method to lock a connection (given by :code:`dsn`) in the specified
        configuration file. Automatically opens the file and writes to it before
        closing.

        :param str conf: The configuration file to modify
        :param str dsn: The name of the connection to lock
        :raises `giraffez.errors.ConfigurationError`: if the connection does not exist
        """
        with Config(conf, "w", key) as c:
            connection = c.get_connection(dsn)
            if not connection:
                raise ConfigurationError("Unable to lock connection")
            if dsn is None:
                dsn = c.settings["connections"]["default"]
            value = "connections.{}.lock".format(dsn)
            lock = c.get_value("connections.{}.lock".format(dsn), default=0)
            if lock >= 2:
                raise ConnectionLock(dsn)
            lock += 1
            c.set_value("connections.{}.lock".format(dsn), lock)
            c.write()

    def read(self):
        with open(self._config_file, "rb") as f:
            return yaml.load(f)

    def reload(self):
        """
        Re-open and read settings from file.
        """
        self.settings = self.read()

    def set_value(self, key, value):
        """
        Set a value within the configuration based on its key. The key
        may be nested, any nested levels that do not exist prior to the final
        segment of the key path will be created.
        *Note*: In order to write changes to the file, ensure that
        :meth:`~giraffez.config.Config.write` is called prior to exit.

        :param str key: A path to the value destination, with nested levels joined by '.'
        :param value: Value to set at the given key, can be any value that is
            YAML serializeable.
        """
        if key.endswith("."):
            key = key[:-1]
        path = key.split(".")
        curr = self.settings
        for p in path[:-1]:
            if p not in curr:
                curr[p] = {}
            curr = curr[p]
        if not isinstance(curr, dict):
            raise ConfigurationError("Cannot set nested key '{}' in configuration value '{}' (destination is not a dictionary).".format(path[-1], key))
        value = self.encrypt(value, path)
        if value in {'true', 'True'}:
            value = True
        if value in {'false', 'False'}:
            value = False
        curr[path[-1]] = value

    @classmethod
    def unlock_connection(cls, conf, dsn, key=None):
        """
        A class method to unlock a connection (given by :code:`dsn`) in the specified
        configuration file. Automatically opens the file and writes to it before
        closing.

        :param str conf: The configuration file to modify
        :param str dsn: The name of the connection to unlock
        :raises `giraffez.errors.ConfigurationError`: if the connection does not exist
        """
        with Config(conf, "w", key) as c:
            connection = c.connections.get(dsn, None)
            if not connection:
                raise ConfigurationError("Unable to unlock connection")
            if dsn is None:
                dsn = c.settings["connections"]["default"]
            if connection.get("lock", None) is None:
                raise GiraffeError("Connection '{}' is not locked.".format(dsn))
            c.unset_value("connections.{}.lock".format(dsn))
            c.write()

    def unset_value(self, key):
        """
        Remove a value at the given key -- and any nested values --
        from the configuration.
        *Note*: In order to write changes to the file, ensure that
        :meth:`~giraffez.config.Config.write` is called prior to exit.

        :param str key: A path to the value destination, with nested levels joined by '.'
        :raises `giraffez.errors.ConfigurationError`: if the key specifies an invalid path, or does not exist
        """
        if key.endswith("."):
            key = key[:-1]
        path = key.split(".")
        curr = self.settings
        for p in path[:-1]:
            if p not in curr:
                raise ConfigurationError("Cannot unset '{}', nested key '{}' does not exist.".format(key, p))
            curr = curr[p]
        if not isinstance(curr, dict):
            raise ConfigurationError("Cannot unset nested key '{}' in configuration value '{}'.".format(path[-1], key))
        if path[-1] not in curr:
            raise ConfigurationError("Cannot unset '{}', nested key '{}' does not exist.".format(key, path[-1]))
        del curr[path[-1]]

    def write(self, settings=None):
        """
        Save the current configuration to its file (as given by :code:`self._config_file`).
        Optionally, settings may be passed in to override the current settings before
        writing. Returns :code:`None` if the file could not be written to, either due to
        permissions, or if the :class:`~giraffez.config.Config` object has the :code:`mode`
        'r'.

        :param dict settings: Defaults to :code:`None`, if not :code:`None` this will replace
            `self.settings` prior to writing to the file
        """
        if "r" in self.mode:
            raise ConfigReadOnly("Cannot write Config while in 'r' mode")
        try:
            if settings:
                self.settings = settings
            with open(self._config_file, "w") as f:
                f.write(repr(self))
                return repr(self)
        except OSError:
            return None

    @classmethod
    def write_default(self, conf=None):
        """
        A class method to write a default configuration file structure to a file.
        Note that the contents of the file will be overwritten if it already exists.

        :param str conf: The name of the file to write to. Defaults to :code:`None`, for ~/.girafferc
        :return: The content written to the file
        :rtype: str
        """
        if conf is None:
            conf = home_file(".girafferc")
        contents = yaml.dump(default_config, default_flow_style=False)
        with open(conf, "w") as f:
            f.write(contents)
        os.chmod(conf, 0o600)
        return contents

    def __repr__(self):
        return yaml.dump(self.settings, default_flow_style=False)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
import warnings

TPT_NOT_FOUND = False
try:
    from . import _tpt
except ImportError:
    TPT_NOT_FOUND = True

from .cmd import *

from .connection import *
from .config import *
from .constants import *
from .core import *
from .encoders import *
from .errors import *
from .fmt import *
from .io import *
from .logging import *
from .types import *
from .utils import *


__all__ = ['TeradataMLoad']


class TeradataMLoad(Connection):
    """
    The class for using the TPT API's UPDATE (MLoad) driver to insert a
    large (> ~100k rows) amount of data into an existing Teradata table.

    Exposed under the alias :class:`giraffez.MLoad`.

    :param str table: The name of the target table for loading.
    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE`, :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE`, and :code:`giraffez.DEBUG`
    :param str config: Specify an alternate configuration file to be read from,
        when previous paramaters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True`,
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>`, changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.errors.TeradataError`: if the connection cannot be established

    If the target table is currently under an MLoad lock (such as if the
    previous operation failed), a :code:`release mload` statement will be
    executed on the table, and the load job will be re-attempted.

    Meant to be used, where possible, with python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete.
    """
    checkpoint_interval = DEFAULT_CHECKPOINT_INTERVAL
    _valid_encodings = {"text", "archive"}

    def __init__(self, table=None, host=None, username=None, password=None, log_level=INFO,
            config=None, key_file=None, dsn=None, protect=False):
        if TPT_NOT_FOUND:
            raise TeradataPTAPINotFound(TeradataPTAPINotFound.__doc__.rstrip())
        super(TeradataMLoad, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect, mload_session=False)

        #: Attributes used with property getter/setters
        self._columns = None
        self._table_name = None
        self._encoding = None

        #: Flags tracking the important stages of MultiLoad to ensure
        #: that the job is shutdown smoothly (if possible).
        self.initiated = False
        self.end_acquisition = False

        self.applied_count = 0
        self.error_count = 0
        self.processor = None

        self.allow_precision_loss = False
        self.encoding = DEFAULT_ENCODING
        if table is not None:
            self.table = table

    def _apply_rows(self):
        log.info("MLoad", "Beginning apply phase ...")
        self.load.apply_rows()
        self._update_apply_count()
        self._update_error_count()
        log.info("MLoad", "Apply phase ended.")

    def _update_apply_count(self):
        data = self.load.get_event(TD_Evt_RowCounts64)
        if data is None:
            log.info("MLoad", "Update apply row count failed.")
            return
        recv, sent, applied = struct.unpack("QQQ", data)
        log.debug("Debug[2]", "Event[RowCounts64]: r:{}, s:{}, a:{}".format(recv, sent, applied))
        self.applied_count = applied

    def _update_error_count(self):
        data = self.load.get_event(TD_Evt_ErrorTable2, 1)
        if data is None:
            log.info("MLoad", "Update error row count failed.")
            return
        count = struct.unpack("I", data)[0]
        log.debug("Debug[2]", "Event[ErrorTable2]: c:{}".format(count))
        self.error_count = count

    def _end_acquisition(self):
        log.info("MLoad", "Ending acquisition phase ...")
        self.load.end_acquisition()
        self.end_acquisition = True
        log.info("MLoad", "Acquisition phase ended.")

    def _connect(self, host, username, password):
        self.cmd = TeradataCmd(host, username, password, log_level=log.level, mload_session=True)
        self.load = _tpt.Load(host, username, password)
        self._handle_error()

    def _handle_error(self, status=None):
        if status is None:
            status = self.status
        if status >= TD_ERROR:
            self.load.get_error_info()
            raise MultiLoadError("{}: {}".format(status, self.load.error_message()))
        return False

    def _initiate(self):
        if not self.table:
            raise GiraffeError("Table must be set prior to initiating.")
        if self.columns is None:
            raise GiraffeError("Columns must be set prior to initiating.")
        if self.initiated:
            raise GiraffeError("Already initiated connection.")
        title, version = get_version_info()
        query_band = "UTILITYNAME={};VERSION={};".format(title, version)
        self.load.add_attribute(TD_QUERY_BAND_SESS_INFO, query_band)
        dml = "insert into {} ({}) values ({});".format(self.table,
            ",".join(['"{}"'.format(f) for f in self.columns.names]),
            ",".join(":{}".format(f) for f in self.columns.safe_names))
        log.verbose("MLoad", "DML: {}".format(dml))
        def _initiate():
            log.info("MLoad", "Initiating Teradata PT request (awaiting server)  ...")
            self.load.initiate(self.columns.tuples(), dml, MARK_DUPLICATE_ROWS)
            self._handle_error()
            log.info("MLoad", "Teradata PT request accepted.")
        try:
            _initiate()
        except (MultiLoadLocked, TransactionAborted) as error:
            log.info("MLoad", "Teradata PT request failed with code '{}'.".format(self.status))
            self.release()
            _initiate()
        if self.processor is None:
            self.processor = pipeline([
                python_to_teradata(self.columns)
            ])
        self.initiated = True

    def checkpoint(self):
        """
        Execute a checkpoint while loading rows. Called automatically
        when loading from a file. Updates the exit code of the driver to
        reflect errors.
        """
        checkpoint_status = self.load.checkpoint()
        if checkpoint_status >= TD_ERROR:
            status_message = MESSAGES.get(int(checkpoint_status), None)
            error_table = [("Error Code", "Error Description")]
            error_table.append((checkpoint_status, status_message))
            log.info("\r{}".format(format_table(error_table)))
        return checkpoint_status

    def cleanup(self):
        """
        Drops any existing work tables, as returned by
        :meth:`~giraffez.mload.TeradataMLoad.tables`.

        :raises `giraffez.errors.TeradataError`: if a Teradata error ocurred
        """
        for table in filter(lambda x: self.cmd.exists(x, silent=(log.level < DEBUG)), self.tables):
            log.info("MLoad", "Dropping table '{}'...".format(table))
            self.cmd.drop_table(table, silent=True)

    def close(self):
        log.info("MLoad", "Closing Teradata PT connection ...")
        if not self.end_acquisition and self.initiated:
            log.info("MLoad", "Acquisition phase was not called before closing.")
            self._end_acquisition()
        self.load.close()
        self.cmd.close()
        log.info("MLoad", "Teradata PT request complete.")

    @property
    def columns(self):
        """
        The list of columns in use.

        :getter: Return the list of columns in use.
        :setter: Set the columns to be loaded into, as well as their order. If
            loading from a file, these will be determined from the file header.
            Not necessary if you are loading into all columns, in the original
            order. The value must be a :code:`list` of names in the order that
            the fields of data will be presented in each row.

            Raises :class:`~giraffez.errors.GiraffeError` if :code:`field_names`
            is not a :code:`list`.

            Raises :class:`~giraffez.errors.GiraffeError` if the target table
            has not been set.
        :type: :class:`~giraffez.types.Columns`
        """
        return self._columns

    @columns.setter
    def columns(self, field_names):
        if not isinstance(field_names, list):
            raise GiraffeError("Must set .columns property as type <List>")
        if not self.table:
            raise GiraffeError("Table name not set")
        self._columns.set_filter(field_names)

    @property
    def encoding(self):
        """
        The encoding of the file being loaded.

        :getter: Returns the name of the encoding being used.
        :setter: Set the encoding of the input file if not specified to the
            constructor of the instance. Accepted values are :code:`'text'`
            or :code:`'archive'`.

            Raises :class:`~giraffez.errors.GiraffeError` if :code:`enc` is not one
            of the above.
        :type: str
        """
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        if value not in self._valid_encodings:
            raise GiraffeError("{} is not a valid encoding type".format(value))
        self._encoding = value

    @property
    def exit_code(self):
        data = self.load.get_event(TD_Evt_ExitCode)
        if data is None:
            log.info("MLoad", "Update exit code failed.")
            return
        return struct.unpack("h", data)[0]

    def finish(self):
        """
        Finishes the load job. **Called automatically when loading from a
        file, but must be called at the end of the job after loading rows
        individually.**

        :return: The exit code returned when applying rows to the table
        """
        checkpoint_status = self.checkpoint()
        self._end_acquisition()
        self._apply_rows()
        exit_code = self.exit_code
        if exit_code != 0:
            self._handle_error()
        return exit_code

    def from_file(self, filename, table=None, null=DEFAULT_NULL, delimiter=None, panic=True, quotechar='"'):
        """
        Load from a file into the target table, handling each step of the
        load process.

        Can load from text files, and properly formatted giraffez archive
        files. In both cases, if Gzip compression is detected the file will be
        decompressed while reading and handled appropriately. The encoding is
        determined automatically by the contents of the file.

        It is not necessary to set the columns in use prior to loading from a file.
        In the case of a text file, the header is used to determine column names
        and their order. Valid delimiters include '|', ',', and '\\t' (tab). When
        loading an archive file, the column information is decoded alongside the data.

        :param str filename: The location of the file to be loaded
        :param str table: The name of the target table, if it was not specified
            to the constructor for the isntance
        :param str null: The string that indicates a null value in the rows being
            inserted from a file. Defaults to 'NULL'
        :param str delimiter: When loading a file, indicates that fields are
            separated by this delimiter. Defaults to :code:`None`, which causes the
            delimiter to be determined from the header of the file. In most
            cases, this behavior is sufficient
        :param str quotechar: The character used to quote fields containing special characters,
            like the delimiter.
        :param bool panic: If :code:`True`, when an error is encountered it will be
            raised. Otherwise, the error will be logged and :code:`self.error_count`
            is incremented.
        :return: The output of the call to
            :meth:`~giraffez.mload.TeradataMLoad.finish`
        :raises `giraffez.errors.GiraffeError`: if table was not set and :code:`table`
            is :code:`None`, or if a Teradata error ocurred while retrieving table info.
        :raises `giraffez.errors.GiraffeEncodeError`: if :code:`panic` is :code:`True` and there
            are format errors in the row values.
        """
        if not self.table:
            if not table:
                raise GiraffeError("Table must be set or specified to load a file.")
            self.table = table

        with Reader(filename, delimiter=delimiter, quotechar=quotechar) as f:
            self.columns = f.header

            if isinstance(f, FileReader):
                self.processor = pipeline([
                    null_handler(null),
                    python_to_teradata(self.columns, self.allow_precision_loss)
                ])
            else:
                self.processor = lambda s: s

            self._initiate()
            i = 0
            for i, line in enumerate(f, 1):
                self.load_row(line, panic=panic)
                if i % self.checkpoint_interval == 1:
                    log.info("\rMLoad", "Processed {} rows".format(i), console=True)
                    checkpoint_status = self.checkpoint()
                    exit_code = self.exit_code
                    if exit_code != 0:
                        return exit_code
            log.info("\rMLoad", "Processed {} rows".format(i))
            return self.finish()

    def load_row(self, items, panic=True):
        """
        Load a single row into the target table.

        :param list items: A list of values in the row corresponding to the
            fields specified by :code:`self.columns`
        :param bool panic: If :code:`True`, when an error is encountered it will be
            raised. Otherwise, the error will be logged and :code:`self.error_count`
            is incremented.
        :raises `giraffez.errors.GiraffeEncodeError`: if :code:`panic` is :code:`True` and there
            are format errors in the row values.
        :raises `giraffez.errors.GiraffeError`: if table name is not set.
        :raises `giraffez.errors.TeradataError`: if there is a problem
            connecting to Teradata.
        """
        if not self.initiated:
            self._initiate()
        try:
            data = self.processor(items)
            row_status = self.load.put_row(data)
            if not self._handle_error(row_status):
                self.applied_count += 1
        except GiraffeEncodeError as error:
            self.error_count += 1
            if panic:
                raise error
            log.info("MLoad", error)

    def release(self):
        """
        Attempt release of target mload table.

        :raises `giraffez.errors.GiraffeError`: if table was not set by
            the constructor, the :code:`TeradataMLoad.table`, or
            :meth:`~giraffez.mload.TeradataMLoad.from_file`.
        """
        if self.table is None:
            raise GiraffeError("Cannot release. Target table has not been set.")
        log.info("MLoad", "Attempting release for table {}".format(self.table))
        self.cmd.release_mload(self.table, silent=True)

    @property
    def status(self):
        return self.load.status()

    @property
    def tables(self):
        """
        The names of the work tables used for loading.

        :return: A list of four tables, each the name of the target table
            with the added suffixes, "_wt", "_log", "_e1", and "_e2"
        :raises `giraffez.errors.GiraffeError`: if table was not set by
            the constructor, the :code:`TeradataMLoad.table`, or
            :meth:`~giraffez.mload.TeradataMLoad.from_file`.
        """
        if self.table is None:
            raise GiraffeError("Target table has not been set.")
        return [
            "{}_wt".format(self.table),
            "{}_log".format(self.table),
            "{}_e1".format(self.table),
            "{}_e2".format(self.table),
        ]

    @property
    def table(self):
        """
        The name of the target table.

        :getter: Returns the name of the target table, or :code:`None` if it
            has not been set.
        :setter: Set the name of the target table, if the table name
            was not given to the constructor of the
            :class:`~giraffez.mload.TeradataMLoad` instance or
            :meth:`~giraffez.mload.TeradataMLoad.from_file`. The value given
            must include all qualifiers such as database name.

            Raises :class:`~giraffez.errors.GiraffeError` if the MLoad connection has
            already been initiated, or the :class:`~giraffez.cmd.TeradataCmd` connection cannot
            be established.

            Raises :class:`~giraffez.errors.TeradataError` if the column data could not be
            retrieved from Teradata
        :type: str
        """
        return self._table_name

    @table.setter
    def table(self, table_name):
        if self.initiated:
            raise GiraffeError("Cannot reuse MLoad context for more than one table")
        self._table_name = table_name
        self.load.set_table(table_name)
        try:
            log.info("MLoad", "Requesting column info for table '{}' ...".format(self.table))
            self._columns = self.cmd.get_columns(self.table, silent=True)
        except MultiLoadTableExists as error:
            log.info("MLoad", "Teradata PT request failed with code '{}'.".format(error.code))
            self.release()
            self._columns = self.cmd.get_columns(self.table, silent=True)

    @property
    def total_count(self):
        """
        The number of rows applied, plus the number of rows in error.
        """
        return self.applied_count + self.error_count



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import os
import sys
import logging
import multiprocessing

from .constants import *
from .errors import *

from ._compat import *


__all__ = ['colors', 'log', 'setup_logging']


logging_level_map = {
    SILENCE: logging.NOTSET,
    INFO: logging.INFO,
    VERBOSE: logging.DEBUG,
    DEBUG: logging.DEBUG
}

def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper

@run_once
def setup_logging():
    l = logging.getLogger("giraffez.console_logger")
    l.addHandler(logging.StreamHandler(sys.stderr))
    l.setLevel(logging.DEBUG)

logging_logger = logging.getLogger("giraffez.console_logger")

CONSOLE_HEADER = "\033[95m"
CONSOLE_BLUE = "\033[94m"
CONSOLE_GREEN = "\033[92m"
CONSOLE_GRAY = "\033[37m"
CONSOLE_WHITE = "\033[97m"
CONSOLE_WARNING = "\033[93m"
CONSOLE_ENDC = "\033[0m"
CONSOLE_FAIL = "\033[91m"
CONSOLE_BOLD = "\033[1m"
CONSOLE_UNDERLINE = "\033[4m" 
CONSOLE_DISABLED = "\033[38;5;240m" 


class colors(object):
    is_colorful = True

    @classmethod
    def colorize(cls, color, s):
        if colors.is_colorful:
            return "{}{}\033[0m".format(color, s)
        return s

    @classmethod
    def blue(cls, s):
        return cls.colorize(CONSOLE_BLUE, s)

    @classmethod
    def green(cls, s):
        return cls.colorize(CONSOLE_GREEN, s)

    @classmethod
    def gray(cls, s):
        return cls.colorize(CONSOLE_GRAY, s)

    @classmethod
    def white(cls, s):
        return cls.colorize(CONSOLE_WHITE, s)

    @classmethod
    def disabled(cls, s):
        return cls.colorize(CONSOLE_DISABLED, s)

    @classmethod
    def bold(cls, s):
        return cls.colorize(CONSOLE_BOLD, s)

    @classmethod
    def fail(cls, s):
        return cls.colorize(CONSOLE_FAIL, s)


def highlight(s):
    return colors.white(colors.bold("{}: ".format(s)))


class Logger(object):
    def __init__(self, fd=None, level=INFO):
        self.fd = fd
        self.level = level

    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, v):
        if v not in [SILENCE, INFO, VERBOSE, DEBUG]:
            raise GiraffeError("Invalid log level set")
        self._level = v

    def _write(self, args, level=None, console=False):
        if level is not None:
            if self.level < level:
                return
        args = list(args)
        if len(args) == 0:
            return
        elif len(args) == 1:
            category = None
        else:
            category = args.pop(0)
        if category is not None:
            category = highlight(category)
        if isinstance(args[0], Exception):
            args = [repr(arg) for arg in args]
        else:
            args = [arg if isinstance(arg, basestring) else str(arg) for arg in args]
        logging_level = logging_level_map.get(level, logging.INFO)
        msg = "".join([str(m) for m in args])
        if category:
            msg = "{}{}".format(category, msg)
        if console:
            sys.stderr.write(msg)
        else:
            logging_logger.log(logging_level, str(msg))
        return self

    def write(self, *args, **kwargs):
        return self._write(args, level=None, console=kwargs.get("console", False))

    def info(self, *args, **kwargs):
        return self._write(args, level=INFO, console=kwargs.get("console", False))

    def verbose(self, *args, **kwargs):
        return self._write(args, level=VERBOSE, console=kwargs.get("console", False))

    def debug(self, *args, **kwargs):
        return self._write(args, level=DEBUG, console=kwargs.get("console", False))

    def fatal(self, msg=""):
        self._write(["System", msg])
        sys.exit(1)


class LockingLogger(Logger):
    def __init__(self, fd=None, level=INFO):
        super(LockingLogger, self).__init__(fd, level)
        self.lock = multiprocessing.Lock()

    def _write(self, args, level=None, console=True):
        with self.lock:
            super(LockingLogger, self)._write(args, level=level, console=console)


log = LockingLogger(sys.stderr, level=INFO)



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains all constants used by giraffez, for settings and for
communicating with Teradata.
"""

ascii_giraffe = """
               ._ O O
               \_`-\|_
            ,""       \       giraffez
          ,":::  |   o o.
        ,"''''  ,-\__    `.
      ,".   .::/     `--._;)
    ,":::  :::/           U
  ,".::::.  :/
"""

GIRAFFE_MAGIC = b"\x5e\x1f\x1e"
GZIP_MAGIC = b"\x1f\x8b"

SUCCESS = 0
FAILURE = 1

SILENCE = -1
INFO    = 0
VERBOSE = 1
DEBUG   = 2

CLI_BLOCK_SIZE = 64260
MLOAD_THRESHOLD = 100000
DEFAULT_NULL = "NULL"
DEFAULT_ENCODING = "text"
DEFAULT_DELIMITER = "|"
DEFAULT_CHECKPOINT_INTERVAL = 50000

### Teradata

# TPT types
TD_NONE                         = 0
TD_INTEGER                      = 1
TD_SMALLINT                     = 2
TD_FLOAT                        = 3
TD_DECIMAL                      = 4
TD_CHAR                         = 5
TD_BYTEINT                      = 6
TD_VARCHAR                      = 7
TD_LONGVARCHAR                  = 8
TD_BYTE                         = 9
TD_VARBYTE                      = 10
TD_DATE                         = 11
TD_GRAPHIC                      = 12
TD_VARGRAPHIC                   = 13
TD_LONGVARGRAPHIC               = 14
TD_DATE_ANSI                    = 15
TD_TIME                         = 35
TD_TIME_WITHTIMEZONE            = 36
TD_BIGINT                       = 37
TD_BLOB                         = 39
TD_CLOB                         = 40
TD_BLOB_AS_DEFERRED_BY_NAME     = 41
TD_CLOB_AS_DEFERRED_BY_NAME     = 42
TD_TIMESTAMP                    = 43
TD_TIMESTAMP_WITHTIMEZONE       = 44
TD_INTERVAL_YEAR                = 45
TD_INTERVAL_YEARTOMONTH         = 46
TD_INTERVAL_MONTH               = 47
TD_INTERVAL_DAY                 = 48
TD_INTERVAL_DAYTOHOUR           = 49
TD_INTERVAL_DAYTOMINUTE         = 50
TD_INTERVAL_DAYTOSECOND         = 51
TD_INTERVAL_HOUR                = 52
TD_INTERVAL_HOURTOMINUTE        = 53
TD_INTERVAL_HOURTOSECOND        = 54
TD_INTERVAL_MINUTE              = 55
TD_INTERVAL_MINUTETOSECOND      = 56
TD_INTERVAL_SECOND              = 57
TD_PERIOD_DATE                  = 58
TD_PERIOD_TIME                  = 59
TD_PERIOD_TIME_TZ               = 60
TD_PERIOD_TS                    = 61
TD_PERIOD_TS_TZ                 = 62
TD_NUMBER                       = 63


#CLIv2 types
BLOB_NN                         = 400
BLOB_N                          = 401
BLOB_AS_DEFERRED_NN             = 404
BLOB_AS_DEFERRED_N              = 405
BLOB_AS_LOCATOR_NN              = 408
BLOB_AS_LOCATOR_N               = 409
BLOB_AS_DEFERRED_NAME_NN        = 412
BLOB_AS_DEFERRED_NAME_N         = 413
CLOB_NN                         = 416
CLOB_N                          = 417
CLOB_AS_DEFERRED_NN             = 420
CLOB_AS_DEFERRED_N              = 421
CLOB_AS_LOCATOR_NN              = 424
CLOB_AS_LOCATOR_N               = 425
UDT_NN                          = 432
UDT_N                           = 433
DISTINCT_UDT_NN                 = 436
DISTINCT_UDT_N                  = 437
STRUCT_UDT_NN                   = 440
STRUCT_UDT_N                    = 441
VARCHAR_NN                      = 448
VARCHAR_N                       = 449
CHAR_NN                         = 452
CHAR_N                          = 453
LONG_VARCHAR_NN                 = 456
LONG_VARCHAR_N                  = 457
VARGRAPHIC_NN                   = 464
VARGRAPHIC_N                    = 465
GRAPHIC_NN                      = 468
GRAPHIC_N                       = 469
LONG_VARGRAPHIC_NN              = 472
LONG_VARGRAPHIC_N               = 473
FLOAT_NN                        = 480
FLOAT_N                         = 481
DECIMAL_NN                      = 484
DECIMAL_N                       = 485
INTEGER_NN                      = 496
INTEGER_N                       = 497
SMALLINT_NN                     = 500
SMALLINT_N                      = 501
ARRAY_1D_NN                     = 504
ARRAY_1D_N                      = 505
ARRAY_ND_NN                     = 508
ARRAY_ND_N                      = 509
BIGINT_NN                       = 600
BIGINT_N                        = 601
VARBYTE_NN                      = 688
VARBYTE_N                       = 689
BYTE_NN                         = 692
BYTE_N                          = 693
LONG_VARBYTE_NN                 = 696
LONG_VARBYTE_N                  = 697
DATE_NNA                        = 748
DATE_NA                         = 749
DATE_NN                         = 752
DATE_N                          = 753
BYTEINT_NN                      = 756
BYTEINT_N                       = 757
TIME_NN                         = 760
TIME_N                          = 761
TIMESTAMP_NN                    = 764
TIMESTAMP_N                     = 765
TIME_NNZ                        = 768
TIME_NZ                         = 769
TIMESTAMP_NNZ                   = 772
TIMESTAMP_NZ                    = 773
INTERVAL_YEAR_NN                = 776
INTERVAL_YEAR_N                 = 777
INTERVAL_YEAR_TO_MONTH_NN       = 780
INTERVAL_YEAR_TO_MONTH_N        = 781
INTERVAL_MONTH_NN               = 784
INTERVAL_MONTH_N                = 785
INTERVAL_DAY_NN                 = 788
INTERVAL_DAY_N                  = 789
INTERVAL_DAY_TO_HOUR_NN         = 792
INTERVAL_DAY_TO_HOUR_N          = 793
INTERVAL_DAY_TO_MINUTE_NN       = 796
INTERVAL_DAY_TO_MINUTE_N        = 797
INTERVAL_DAY_TO_SECOND_NN       = 800
INTERVAL_DAY_TO_SECOND_N        = 801
INTERVAL_HOUR_NN                = 804
INTERVAL_HOUR_N                 = 805
INTERVAL_HOUR_TO_MINUTE_NN      = 808
INTERVAL_HOUR_TO_MINUTE_N       = 809
INTERVAL_HOUR_TO_SECOND_NN      = 812
INTERVAL_HOUR_TO_SECOND_N       = 813
INTERVAL_MINUTE_NN              = 816
INTERVAL_MINUTE_N               = 817
INTERVAL_MINUTE_TO_SECOND_NN    = 820
INTERVAL_MINUTE_TO_SECOND_N     = 821
INTERVAL_SECOND_NN              = 824
INTERVAL_SECOND_N               = 825
PERIOD_DATE_NN                  = 832
PERIOD_DATE_N                   = 833
PERIOD_TIME_NN                  = 836
PERIOD_TIME_N                   = 837
PERIOD_TIME_NNZ                 = 840
PERIOD_TIME_NZ                  = 841
PERIOD_TIMESTAMP_NN             = 844
PERIOD_TIMESTAMP_N              = 845
PERIOD_TIMESTAMP_NNZ            = 848
PERIOD_TIMESTAMP_NZ             = 849
XML_TEXT_NN                     = 852
XML_TEXT_N                      = 853
XML_TEXT_DEFERRED_NN            = 856
XML_TEXT_DEFERRED_N             = 857
XML_TEXT_LOCATOR_NN             = 860
XML_TEXT_LOCATOR_N              = 861

# CLIv2 type code -> name
COL_TYPE_NAMES = {
    BLOB_NN: "BLOB_NN",
    BLOB_N: "BLOB_N",
    BLOB_AS_DEFERRED_NN: "BLOB_AS_DEFERRED_NN",
    BLOB_AS_DEFERRED_N: "BLOB_AS_DEFERRED_N",
    BLOB_AS_LOCATOR_NN: "BLOB_AS_LOCATOR_NN",
    BLOB_AS_LOCATOR_N: "BLOB_AS_LOCATOR_N",
    BLOB_AS_DEFERRED_NAME_NN: "BLOB_AS_DEFERRED_NAME_NN",
    BLOB_AS_DEFERRED_NAME_N: "BLOB_AS_DEFERRED_NAME_N",
    CLOB_NN: "CLOB_NN",
    CLOB_N: "CLOB_N",
    CLOB_AS_DEFERRED_NN: "CLOB_AS_DEFERRED_NN",
    CLOB_AS_DEFERRED_N: "CLOB_AS_DEFERRED_N",
    CLOB_AS_LOCATOR_NN: "CLOB_AS_LOCATOR_NN",
    CLOB_AS_LOCATOR_N: "CLOB_AS_LOCATOR_N",
    UDT_NN: "UDT_NN",
    UDT_N: "UDT_N",
    DISTINCT_UDT_NN: "DISTINCT_UDT_NN",
    DISTINCT_UDT_N: "DISTINCT_UDT_N",
    STRUCT_UDT_NN: "STRUCT_UDT_NN",
    STRUCT_UDT_N: "STRUCT_UDT_N",
    VARCHAR_NN: "VARCHAR_NN",
    VARCHAR_N: "VARCHAR_N",
    CHAR_NN: "CHAR_NN",
    CHAR_N: "CHAR_N",
    LONG_VARCHAR_NN: "LONG_VARCHAR_NN",
    LONG_VARCHAR_N: "LONG_VARCHAR_N",
    VARGRAPHIC_NN: "VARGRAPHIC_NN",
    VARGRAPHIC_N: "VARGRAPHIC_N",
    GRAPHIC_NN: "GRAPHIC_NN",
    GRAPHIC_N: "GRAPHIC_N",
    LONG_VARGRAPHIC_NN: "LONG_VARGRAPHIC_NN",
    LONG_VARGRAPHIC_N: "LONG_VARGRAPHIC_N",
    FLOAT_NN: "FLOAT_NN",
    FLOAT_N: "FLOAT_N",
    DECIMAL_NN: "DECIMAL_NN",
    DECIMAL_N: "DECIMAL_N",
    INTEGER_NN: "INTEGER_NN",
    INTEGER_N: "INTEGER_N",
    SMALLINT_NN: "SMALLINT_NN",
    SMALLINT_N: "SMALLINT_N",
    ARRAY_1D_NN: "ARRAY_1D_NN",
    ARRAY_1D_N: "ARRAY_1D_N",
    ARRAY_ND_NN: "ARRAY_ND_NN",
    ARRAY_ND_N: "ARRAY_ND_N",
    BIGINT_NN: "BIGINT_NN",
    BIGINT_N: "BIGINT_N",
    VARBYTE_NN: "VARBYTE_NN",
    VARBYTE_N: "VARBYTE_N",
    BYTE_NN: "BYTE_NN",
    BYTE_N: "BYTE_N",
    LONG_VARBYTE_NN: "LONG_VARBYTE_NN",
    LONG_VARBYTE_N: "LONG_VARBYTE_N",
    DATE_NNA: "DATE_NNA",
    DATE_NA: "DATE_NA",
    DATE_NN: "DATE_NN",
    DATE_N: "DATE_N",
    BYTEINT_NN: "BYTEINT_NN",
    BYTEINT_N: "BYTEINT_N",
    TIME_NN: "TIME_NN",
    TIME_N: "TIME_N",
    TIMESTAMP_NN: "TIMESTAMP_NN",
    TIMESTAMP_N: "TIMESTAMP_N",
    TIME_NNZ: "TIME_NNZ",
    TIME_NZ: "TIME_NZ",
    TIMESTAMP_NNZ: "TIMESTAMP_NNZ",
    TIMESTAMP_NZ: "TIMESTAMP_NZ",
    INTERVAL_YEAR_NN: "INTERVAL_YEAR_NN",
    INTERVAL_YEAR_N: "INTERVAL_YEAR_N",
    INTERVAL_YEAR_TO_MONTH_NN: "INTERVAL_YEAR_TO_MONTH_NN",
    INTERVAL_YEAR_TO_MONTH_N: "INTERVAL_YEAR_TO_MONTH_N",
    INTERVAL_MONTH_NN: "INTERVAL_MONTH_NN",
    INTERVAL_MONTH_N: "INTERVAL_MONTH_N",
    INTERVAL_DAY_NN: "INTERVAL_DAY_NN",
    INTERVAL_DAY_N: "INTERVAL_DAY_N",
    INTERVAL_DAY_TO_HOUR_NN: "INTERVAL_DAY_TO_HOUR_NN",
    INTERVAL_DAY_TO_HOUR_N: "INTERVAL_DAY_TO_HOUR_N",
    INTERVAL_DAY_TO_MINUTE_NN: "INTERVAL_DAY_TO_MINUTE_NN",
    INTERVAL_DAY_TO_MINUTE_N: "INTERVAL_DAY_TO_MINUTE_N",
    INTERVAL_DAY_TO_SECOND_NN: "INTERVAL_DAY_TO_SECOND_NN",
    INTERVAL_DAY_TO_SECOND_N: "INTERVAL_DAY_TO_SECOND_N",
    INTERVAL_HOUR_NN: "INTERVAL_HOUR_NN",
    INTERVAL_HOUR_N: "INTERVAL_HOUR_N",
    INTERVAL_HOUR_TO_MINUTE_NN: "INTERVAL_HOUR_TO_MINUTE_NN",
    INTERVAL_HOUR_TO_MINUTE_N: "INTERVAL_HOUR_TO_MINUTE_N",
    INTERVAL_HOUR_TO_SECOND_NN: "INTERVAL_HOUR_TO_SECOND_NN",
    INTERVAL_HOUR_TO_SECOND_N: "INTERVAL_HOUR_TO_SECOND_N",
    INTERVAL_MINUTE_NN: "INTERVAL_MINUTE_NN",
    INTERVAL_MINUTE_N: "INTERVAL_MINUTE_N",
    INTERVAL_MINUTE_TO_SECOND_NN: "INTERVAL_MINUTE_TO_SECOND_NN",
    INTERVAL_MINUTE_TO_SECOND_N: "INTERVAL_MINUTE_TO_SECOND_N",
    INTERVAL_SECOND_NN: "INTERVAL_SECOND_NN",
    INTERVAL_SECOND_N: "INTERVAL_SECOND_N",
    PERIOD_DATE_NN: "PERIOD_DATE_NN",
    PERIOD_DATE_N: "PERIOD_DATE_N",
    PERIOD_TIME_NN: "PERIOD_TIME_NN",
    PERIOD_TIME_N: "PERIOD_TIME_N",
    PERIOD_TIME_NNZ: "PERIOD_TIME_NNZ",
    PERIOD_TIME_NZ: "PERIOD_TIME_NZ",
    PERIOD_TIMESTAMP_NN: "PERIOD_TIMESTAMP_NN",
    PERIOD_TIMESTAMP_N: "PERIOD_TIMESTAMP_N",
    PERIOD_TIMESTAMP_NNZ: "PERIOD_TIMESTAMP_NNZ",
    PERIOD_TIMESTAMP_NZ: "PERIOD_TIMESTAMP_NZ",
    XML_TEXT_NN: "XML_TEXT_NN",
    XML_TEXT_N: "XML_TEXT_N",
    XML_TEXT_DEFERRED_NN: "XML_TEXT_DEFERRED_NN",
    XML_TEXT_DEFERRED_N: "XML_TEXT_DEFERRED_N",
    XML_TEXT_LOCATOR_NN: "XML_TEXT_LOCATOR_NN",
    XML_TEXT_LOCATOR_N: "XML_TEXT_LOCATOR_N",
}

NULLABLE_TYPES = {
    BLOB_N,
    BLOB_AS_DEFERRED_N,
    BLOB_AS_LOCATOR_N,
    BLOB_AS_DEFERRED_NAME_N,
    CLOB_N,
    CLOB_AS_DEFERRED_N,
    CLOB_AS_LOCATOR_N,
    UDT_N,
    DISTINCT_UDT_N,
    STRUCT_UDT_N,
    VARCHAR_N,
    CHAR_N,
    LONG_VARCHAR_N,
    VARGRAPHIC_N,
    GRAPHIC_N,
    LONG_VARGRAPHIC_N,
    FLOAT_N,
    DECIMAL_N,
    INTEGER_N,
    SMALLINT_N,
    ARRAY_1D_N,
    ARRAY_ND_N,
    BIGINT_N,
    VARBYTE_N,
    BYTE_N,
    LONG_VARBYTE_N,
    DATE_NA,
    DATE_N,
    BYTEINT_N,
    TIME_N,
    TIMESTAMP_N,
    TIME_NZ,
    TIMESTAMP_NZ,
    INTERVAL_YEAR_N,
    INTERVAL_YEAR_TO_MONTH_N,
    INTERVAL_MONTH_N,
    INTERVAL_DAY_N,
    INTERVAL_DAY_TO_HOUR_N,
    INTERVAL_DAY_TO_MINUTE_N,
    INTERVAL_DAY_TO_SECOND_N,
    INTERVAL_HOUR_N,
    INTERVAL_HOUR_TO_MINUTE_N,
    INTERVAL_HOUR_TO_SECOND_N,
    INTERVAL_MINUTE_N,
    INTERVAL_MINUTE_TO_SECOND_N,
    INTERVAL_SECOND_N,
    PERIOD_DATE_N,
    PERIOD_TIME_N,
    PERIOD_TIME_NZ,
    PERIOD_TIMESTAMP_N,
    PERIOD_TIMESTAMP_NZ,
    XML_TEXT_N,
    XML_TEXT_DEFERRED_N,
    XML_TEXT_LOCATOR_N
}

# Teradata uses space-padded characters representing nulls
STRING_NULLS = {
    CHAR_N,
    CHAR_NN,
    TIME_N,
    TIME_NN,
    DATE_N,
    DATE_NN,
    TIMESTAMP_N,
    TIMESTAMP_NN,
    TIMESTAMP_NZ
}

# Null is 0-length variable header
VARIABLE_NULLS = {
    VARCHAR_N,
    VARCHAR_NN,
    LONG_VARCHAR_N,
    LONG_VARCHAR_NN,
    VARBYTE_N,
    VARBYTE_NN,
    LONG_VARBYTE_N,
    LONG_VARBYTE_NN,
    VARGRAPHIC_NN,
    VARGRAPHIC_N,
    LONG_VARGRAPHIC_NN,
    LONG_VARGRAPHIC_N,
}

BYTE_TYPES = {
    BYTE_N,
    BYTE_NN
}

CHAR_TYPES = {
    CHAR_N,
    CHAR_NN
}

# Numeric types
BYTEINT_TYPES = {
    BYTEINT_N,
    BYTEINT_NN
}

SMALLINT_TYPES = {
    SMALLINT_N,
    SMALLINT_NN
}

INTEGER_TYPES = {
    INTEGER_N,
    INTEGER_NN
}

BIGINT_TYPES = {
    BIGINT_N,
    BIGINT_NN
}

ALL_INTEGER_TYPES = {
    BYTEINT_N,
    BYTEINT_NN,
    SMALLINT_N,
    SMALLINT_NN,
    INTEGER_N,
    INTEGER_NN,
    BIGINT_N,
    BIGINT_NN
}

FLOAT_TYPES = {
    FLOAT_N,
    FLOAT_NN
}

DECIMAL_TYPES = {
    DECIMAL_N,
    DECIMAL_NN
}

VAR_TYPES = {
    VARCHAR_N,
    VARCHAR_NN,
    LONG_VARCHAR_N,
    LONG_VARCHAR_NN,
    VARBYTE_N,
    VARBYTE_NN,
    LONG_VARBYTE_N,
    LONG_VARBYTE_NN,
    VARGRAPHIC_NN,
    VARGRAPHIC_N,
    LONG_VARGRAPHIC_NN,
    LONG_VARGRAPHIC_N,
}

DATE_TYPES = {
    DATE_N,
    DATE_NN
}

TIME_TYPES = {
    TIME_N,
    TIME_NN
}

TIMESTAMP_TYPES = {
    TIMESTAMP_N,
    TIMESTAMP_NN,
    TIMESTAMP_NZ
}

# Event codes
TD_Evt_BufferLayout             = 1
TD_Evt_ConnectStatus            = 2
TD_Evt_RowCounts                = 3
TD_Evt_BlockCount               = 4
TD_Evt_ApplyCount               = 5
TD_Evt_ExportCount              = 6
TD_Evt_ErrorTable1              = 7
TD_Evt_ErrorTable2              = 8
TD_Evt_CLIError                 = 9
TD_Evt_DBSError                 = 10
TD_Evt_CPUTime                  = 11
TD_Evt_LastError                = 12
TD_Evt_ExitCode                 = 13
TD_Evt_RunStats                 = 14
TD_Evt_UserCommand              = 15
TD_Evt_RowsCheckpointed         = 16
TD_Evt_BufferModeSupport        = 17
TD_Evt_MacroNames               = 18
TD_Evt_PackFactor               = 19
TD_Evt_OperVersion              = 20
TD_Evt_RowCounts64              = 21
TD_Evt_RowsCheckpointed64       = 22
TD_Evt_ExportCount64            = 23
TD_Evt_RunStats64               = 24
TD_Evt_ErrorTable1_64           = 25
TD_Evt_ApplyCount64             = 26
TD_Evt_AcquisitionPhaseStats    = 27
TD_Evt_ApplicationPhaseStat     = 28
TD_Evt_LoadPhaseStats           = 29
TD_Evt_ExportPhaseStats         = 30

TD_Evt_Version                  = 101

# Status codes
TD_SUCCESS                      = 0
TD_UNAVAILABLE                  = 4
TD_ERROR                        = 99

# Error codes
TD_ERR_WORK_TABLE_MISSING       = 2583
TD_ERR_TRANS_ABORT              = 2631
TD_ERR_USER_NO_SELECT_ACCESS    = 3523
TD_ERR_OBJECT_NOT_EXIST         = 3807
TD_ERR_OBJECT_NOT_TABLE         = 3853
TD_ERR_OBJECT_NOT_VIEW          = 3854

# CLIv2 Error codes
CLI_ERR_CANNOT_RELEASE_MLOAD    = 2572
CLI_ERR_TABLE_MLOAD_EXISTS      = 2574
CLI_ERR_INVALID_USER            = 8017

# Putting the most common message texts here for helping to debug loads. There
# might be a better way to do this, but may not be worth it to code something
# that looks for and digs through the binary message catalog.
MESSAGES = {
    2620: "The format or data contains a bad character.",
    2655: "Invalid parcel sequences during Load.",
    2665: "Invalid date.",
    2673: "The source parcel length does not match data that was defined.",
    2674: "Precision loss during data conversion.",
    2675: "Numerical overflow occurred during computation.",
    2679: "The format or data contains a bad character.",
    2682: "Precision loss during data conversion.",
    2683: "Numerical overflow occurred during computation.",
    2687: "The format or data contains a bad character.",
    2689: "Field is null for column defined as NOT NULL.",
    2694: "The row is incorrectly formatted.",
    2644: "No more room in database.", # Does NOT populate in the error log(s) (_e1, _e2)
}

# Attributes
TD_SYSTEM_OPERATOR              = 0
TD_USER_NAME                    = 1
TD_USER_PASSWORD                = 2
TD_TDP_ID                       = 3
TD_LOG_TABLE                    = 4
TD_TARGET_TABLE                 = 5
TD_ERROR_TABLE_1                = 6
TD_ERROR_TABLE_2                = 7
TD_ERROR_TABLE                  = 8
TD_BUFFERS                      = 9
TD_BUFFER_SIZE                  = 10
TD_BUFFER_MODE                  = 11
TD_BUFFER_MAX_SIZE              = 12
TD_BUFFER_HEADER_SIZE           = 13
TD_BUFFER_LENGTH_SIZE           = 14
TD_BUFFER_TRAILER_SIZE          = 15
TD_ERROR_LIMIT                  = 16
TD_MAX_SESSIONS                 = 17
TD_MIN_SESSIONS                 = 18
TD_RESTARTMODE                  = 19
TD_INSTANCE_NUM                 = 20
TD_MAX_INSTANCES                = 21
TD_CHARSET                      = 22
TD_TENACITY_HOURS               = 23
TD_TENACITY_SLEEP               = 24
TD_ACCOUNT_ID                   = 25
TD_DATE_FORM                    = 26
TD_NOTIFY_EXIT                  = 27
TD_NOTIFY_EXIT_IS_DLL           = 28
TD_NOTIFY_LEVEL                 = 29
TD_NOTIFY_METHOD                = 30
TD_NOTIFY_STRING                = 31
TD_PRIVATE_LOG_NAME             = 32
TD_TRACE_LEVEL                  = 33
TD_TRACE_OUTPUT                 = 34
TD_WORKINGDATABASE              = 35
TD_PAUSE_ACQ                    = 36
TD_AMP_CHECK                    = 37
TD_DELETE_TASK                  = 38
TD_WORK_TABLE                   = 39
TD_PACK                         = 40
TD_PACKMAXIMUM                  = 41
TD_MACRODATABASE                = 42
TD_ROBUST                       = 43
TD_SELECT_STMT                  = 44
TD_BLOCK_SIZE                   = 45
TD_DATA_ENCRYPTION              = 46
TD_LOGON_MECH                   = 47
TD_LOGON_MECH_DATA              = 48
TD_WILDCARDINSERT               = 49
TD_REPLICATION_OVERRIDE         = 50
TD_MAX_DECIMAL_DIGITS           = 51 
TD_IGNORE_MAX_DECIMAL_DIGITS    = 52
TD_QUERY_BAND_SESS_INFO         = 53
TD_ARRAYSUPPORT                 = 54
TD_MSG_ENCODING                 = 55
TD_APPENDERRORTABLE             = 56
TD_DROPERRORTABLE               = 57
TD_DROPLOGTABLE                 = 58
TD_DROPWORKTABLE                = 59
TD_DROPMACRO                    = 60
TD_OUTLIMIT                     = 61
TD_SPOOLMODE                    = 62
TD_LOGSQL                       = 63
TD_AUTORESTART                  = 64
TD_TMSMSUPPORT                  = 65
TD_RATE                         = 66
TD_PERIODICITY                  = 67
TD_MACROCHARSET                 = 68
TD_RETN_ACT_DATATYPE            = 69

MARK_DUPLICATE_ROWS             = 0
MARK_DUPLICATE_INSERT_ROWS      = 1
MARK_DUPLICATE_UPDATE_ROWS      = 2
MARK_MISSING_ROWS               = 3
MARK_MISSING_UPDATE_ROWS        = 4
MARK_MISSING_DELETE_ROWS        = 5
IGNORE_DUPLICATE_ROWS           = 6
IGNORE_DUPLICATE_INSERT_ROWS    = 7
IGNORE_DUPLICATE_UPDATE_ROWS    = 8
IGNORE_MISSING_ROWS             = 9
IGNORE_MISSING_UPDATE_ROWS      = 10
IGNORE_MISSING_DELETE_ROWS      = 11
INSERT_FOR_MISSING_UPDATE_ROWS  = 12
MARK_EXTRA_ROWS                 = 13
MARK_EXTRA_UPDATE_ROWS          = 14
MARK_EXTRA_DELETE_ROWS          = 15
IGNORE_EXTRA_ROWS               = 16
IGNORE_EXTRA_UPDATE_ROWS        = 17
IGNORE_EXTRA_DELETE_ROWS        = 18

# Type maps between TPT and CLIv2
tpt_type_map = {
    BLOB_NN: TD_BLOB,
    BLOB_N: TD_BLOB,
    BLOB_AS_DEFERRED_NN: TD_BLOB,
    BLOB_AS_DEFERRED_N: TD_BLOB,
    BLOB_AS_LOCATOR_NN: TD_BLOB,
    BLOB_AS_LOCATOR_N: TD_BLOB,
    BLOB_AS_DEFERRED_NAME_NN: TD_BLOB_AS_DEFERRED_BY_NAME,
    BLOB_AS_DEFERRED_NAME_N: TD_BLOB_AS_DEFERRED_BY_NAME,
    CLOB_NN: TD_CLOB,
    CLOB_N: TD_CLOB,
    CLOB_AS_DEFERRED_NN: TD_CLOB_AS_DEFERRED_BY_NAME,
    CLOB_AS_DEFERRED_N: TD_CLOB_AS_DEFERRED_BY_NAME,
    CLOB_AS_LOCATOR_NN: TD_CLOB,
    CLOB_AS_LOCATOR_N: TD_CLOB,
    UDT_NN: TD_CHAR,
    UDT_N: TD_CHAR,
    DISTINCT_UDT_NN: TD_CHAR,
    DISTINCT_UDT_N: TD_CHAR,
    STRUCT_UDT_NN: TD_CHAR,
    STRUCT_UDT_N: TD_CHAR,
    VARCHAR_NN: TD_VARCHAR,
    VARCHAR_N: TD_VARCHAR,
    CHAR_NN: TD_CHAR,
    CHAR_N: TD_CHAR,
    LONG_VARCHAR_NN: TD_LONGVARCHAR,
    LONG_VARCHAR_N: TD_LONGVARCHAR,
    VARGRAPHIC_NN: TD_VARGRAPHIC,
    VARGRAPHIC_N: TD_VARGRAPHIC,
    GRAPHIC_NN: TD_GRAPHIC,
    GRAPHIC_N: TD_GRAPHIC,
    LONG_VARGRAPHIC_NN: TD_LONGVARGRAPHIC,
    LONG_VARGRAPHIC_N: TD_LONGVARGRAPHIC,
    FLOAT_NN: TD_FLOAT,
    FLOAT_N: TD_FLOAT,
    DECIMAL_NN: TD_DECIMAL,
    DECIMAL_N: TD_DECIMAL,
    INTEGER_NN: TD_INTEGER,
    INTEGER_N: TD_INTEGER,
    SMALLINT_NN: TD_SMALLINT,
    SMALLINT_N: TD_SMALLINT,
    ARRAY_1D_NN: TD_CHAR,
    ARRAY_1D_N: TD_CHAR,
    ARRAY_ND_NN: TD_CHAR,
    ARRAY_ND_N: TD_CHAR,
    BIGINT_NN: TD_BIGINT,
    BIGINT_N: TD_BIGINT,
    VARBYTE_NN: TD_VARBYTE,
    VARBYTE_N: TD_VARBYTE,
    BYTE_NN: TD_BYTE,
    BYTE_N: TD_BYTE,
    LONG_VARBYTE_NN: TD_LONGVARCHAR,
    LONG_VARBYTE_N: TD_LONGVARCHAR,
    DATE_NNA: TD_CHAR,
    DATE_NA: TD_CHAR,
    DATE_NN: TD_DATE,
    DATE_N: TD_DATE,
    BYTEINT_NN: TD_BYTEINT,
    BYTEINT_N: TD_BYTEINT,
    TIME_NN: TD_TIME,
    TIME_N: TD_TIME,
    TIMESTAMP_NN: TD_TIMESTAMP,
    TIMESTAMP_N: TD_TIMESTAMP,
    TIME_NNZ: TD_TIME_WITHTIMEZONE,
    TIME_NZ: TD_TIME_WITHTIMEZONE,
    TIMESTAMP_NNZ: TD_TIMESTAMP_WITHTIMEZONE,
    TIMESTAMP_NZ: TD_TIMESTAMP_WITHTIMEZONE,
    INTERVAL_YEAR_NN: TD_INTERVAL_YEAR,
    INTERVAL_YEAR_N: TD_INTERVAL_YEAR,
    INTERVAL_YEAR_TO_MONTH_NN: TD_INTERVAL_YEARTOMONTH,
    INTERVAL_YEAR_TO_MONTH_N: TD_INTERVAL_YEARTOMONTH,
    INTERVAL_MONTH_NN: TD_INTERVAL_MONTH,
    INTERVAL_MONTH_N: TD_INTERVAL_MONTH,
    INTERVAL_DAY_NN: TD_INTERVAL_DAY,
    INTERVAL_DAY_N: TD_INTERVAL_DAY,
    INTERVAL_DAY_TO_HOUR_NN: TD_INTERVAL_DAYTOHOUR,
    INTERVAL_DAY_TO_HOUR_N: TD_INTERVAL_DAYTOHOUR,
    INTERVAL_DAY_TO_MINUTE_NN: TD_INTERVAL_DAYTOMINUTE,
    INTERVAL_DAY_TO_MINUTE_N: TD_INTERVAL_DAYTOMINUTE,
    INTERVAL_DAY_TO_SECOND_NN: TD_INTERVAL_DAYTOSECOND,
    INTERVAL_DAY_TO_SECOND_N: TD_INTERVAL_DAYTOSECOND,
    INTERVAL_HOUR_NN: TD_INTERVAL_HOUR,
    INTERVAL_HOUR_N: TD_INTERVAL_HOUR,
    INTERVAL_HOUR_TO_MINUTE_NN: TD_INTERVAL_HOURTOMINUTE,
    INTERVAL_HOUR_TO_MINUTE_N: TD_INTERVAL_HOURTOMINUTE,
    INTERVAL_HOUR_TO_SECOND_NN: TD_INTERVAL_HOURTOSECOND,
    INTERVAL_HOUR_TO_SECOND_N: TD_INTERVAL_HOURTOSECOND,
    INTERVAL_MINUTE_NN: TD_INTERVAL_MINUTE,
    INTERVAL_MINUTE_N: TD_INTERVAL_MINUTE,
    INTERVAL_MINUTE_TO_SECOND_NN: TD_INTERVAL_MINUTETOSECOND,
    INTERVAL_MINUTE_TO_SECOND_N: TD_INTERVAL_MINUTETOSECOND,
    INTERVAL_SECOND_NN: TD_INTERVAL_SECOND,
    INTERVAL_SECOND_N: TD_INTERVAL_SECOND,
    PERIOD_DATE_NN: TD_PERIOD_DATE,
    PERIOD_DATE_N: TD_PERIOD_DATE,
    PERIOD_TIME_NN: TD_PERIOD_TIME,
    PERIOD_TIME_N: TD_PERIOD_TIME,
    PERIOD_TIME_NNZ: TD_PERIOD_TIME_TZ,
    PERIOD_TIME_NZ: TD_PERIOD_TIME_TZ,
    PERIOD_TIMESTAMP_NN: TD_PERIOD_TS,
    PERIOD_TIMESTAMP_N: TD_PERIOD_TS,
    PERIOD_TIMESTAMP_NNZ: TD_PERIOD_TS_TZ,
    PERIOD_TIMESTAMP_NZ: TD_PERIOD_TS_TZ,
    XML_TEXT_NN: TD_CHAR,
    XML_TEXT_N: TD_CHAR,
    XML_TEXT_DEFERRED_NN: TD_CHAR,
    XML_TEXT_DEFERRED_N: TD_CHAR,
    XML_TEXT_LOCATOR_NN: TD_CHAR,
    XML_TEXT_LOCATOR_N: TD_CHAR,
}


cmd_type_map = {
    TD_INTEGER: INTEGER_NN,
    TD_SMALLINT: SMALLINT_NN,
    TD_FLOAT: FLOAT_NN,
    TD_DECIMAL: DECIMAL_NN,
    TD_CHAR: CHAR_NN,
    TD_BYTEINT: BYTEINT_NN,
    TD_VARCHAR: VARCHAR_NN,
    TD_LONGVARCHAR: LONG_VARCHAR_NN,
    TD_BYTE: BYTE_NN,
    TD_VARBYTE: VARBYTE_NN,
    TD_DATE: DATE_NN,
    TD_GRAPHIC: GRAPHIC_NN,
    TD_VARGRAPHIC: VARGRAPHIC_NN,
    TD_LONGVARGRAPHIC: LONG_VARGRAPHIC_NN,
    TD_DATE_ANSI: DATE_NNA,
    TD_TIME: TIME_NN,
    TD_TIME_WITHTIMEZONE: TIME_NNZ,
    TD_BIGINT: BIGINT_NN,
    TD_BLOB: BLOB_NN,
    TD_CLOB: CLOB_NN,
    TD_BLOB_AS_DEFERRED_BY_NAME: BLOB_AS_DEFERRED_NAME_NN,
    TD_CLOB_AS_DEFERRED_BY_NAME: CLOB_AS_DEFERRED_NN,
    TD_TIMESTAMP: TIMESTAMP_NN,
    TD_TIMESTAMP_WITHTIMEZONE: TIMESTAMP_NNZ,
    TD_INTERVAL_YEAR: INTERVAL_YEAR_NN,
    TD_INTERVAL_YEARTOMONTH: INTERVAL_YEAR_TO_MONTH_NN,
    TD_INTERVAL_MONTH: INTERVAL_MONTH_NN,
    TD_INTERVAL_DAY: INTERVAL_DAY_NN,
    TD_INTERVAL_DAYTOHOUR: INTERVAL_DAY_TO_HOUR_NN,
    TD_INTERVAL_DAYTOMINUTE: INTERVAL_DAY_TO_MINUTE_NN,
    TD_INTERVAL_DAYTOSECOND: INTERVAL_DAY_TO_SECOND_NN,
    TD_INTERVAL_HOUR: INTERVAL_HOUR_NN,
    TD_INTERVAL_HOURTOMINUTE: INTERVAL_HOUR_TO_MINUTE_NN,
    TD_INTERVAL_HOURTOSECOND: INTERVAL_HOUR_TO_SECOND_NN,
    TD_INTERVAL_MINUTE: INTERVAL_MINUTE_NN,
    TD_INTERVAL_MINUTETOSECOND: INTERVAL_MINUTE_TO_SECOND_NN,
    TD_INTERVAL_SECOND: INTERVAL_SECOND_NN,
    TD_PERIOD_DATE: PERIOD_DATE_NN,
    TD_PERIOD_TIME: PERIOD_TIME_NN,
    TD_PERIOD_TIME_TZ: PERIOD_TIME_NNZ,
    TD_PERIOD_TS: PERIOD_TIMESTAMP_NN,
    TD_PERIOD_TS_TZ: PERIOD_TIMESTAMP_NNZ,
    TD_NUMBER: INTEGER_NN, # not sure on this one
}

# giraffez data types
# These types are used internally for the purpose of reducing
# the complexity of the underlying encoder type switch while
# not losing any Teradata specific type information.

# GD_DEFAULT will always be equivalent to GD_CHAR in that it advances
# the buffer and then stores that value by the column length (fixed). 
# It will be used when the exact method for handling the type is not
# known and fixed-length is presumed, unless it is discovered that
# type is not fixed-length.  This is done because some less common
# types should be noted but may not be explicitly handled as of yet.
GD_DEFAULT          = 0
GD_BYTEINT          = 1
GD_SMALLINT         = 2
GD_INTEGER          = 3
GD_BIGINT           = 4
GD_FLOAT            = 5
GD_DECIMAL          = 6
GD_CHAR             = 7
GD_VARCHAR          = 8
GD_DATE             = 9

gd_type_map = {
    BLOB_NN: GD_DEFAULT,
    BLOB_N: GD_DEFAULT,
    BLOB_AS_DEFERRED_NN: GD_DEFAULT,
    BLOB_AS_DEFERRED_N: GD_DEFAULT,
    BLOB_AS_LOCATOR_NN: GD_DEFAULT,
    BLOB_AS_LOCATOR_N: GD_DEFAULT,
    BLOB_AS_DEFERRED_NAME_NN: GD_DEFAULT,
    BLOB_AS_DEFERRED_NAME_N: GD_DEFAULT,
    CLOB_NN: GD_DEFAULT,
    CLOB_N: GD_DEFAULT,
    CLOB_AS_DEFERRED_NN: GD_DEFAULT,
    CLOB_AS_DEFERRED_N: GD_DEFAULT,
    CLOB_AS_LOCATOR_NN: GD_DEFAULT,
    CLOB_AS_LOCATOR_N: GD_DEFAULT,
    UDT_NN: GD_DEFAULT,
    UDT_N: GD_DEFAULT,
    DISTINCT_UDT_NN: GD_DEFAULT,
    DISTINCT_UDT_N: GD_DEFAULT,
    STRUCT_UDT_NN: GD_DEFAULT,
    STRUCT_UDT_N: GD_DEFAULT,
    VARCHAR_NN: GD_VARCHAR,
    VARCHAR_N: GD_VARCHAR,
    CHAR_NN: GD_CHAR,
    CHAR_N: GD_CHAR,
    LONG_VARCHAR_NN: GD_VARCHAR,
    LONG_VARCHAR_N: GD_VARCHAR,
    VARGRAPHIC_NN: GD_VARCHAR,
    VARGRAPHIC_N: GD_VARCHAR,
    GRAPHIC_NN: GD_DEFAULT,
    GRAPHIC_N: GD_DEFAULT,
    LONG_VARGRAPHIC_NN: GD_VARCHAR,
    LONG_VARGRAPHIC_N: GD_VARCHAR,
    FLOAT_NN: GD_FLOAT,
    FLOAT_N: GD_FLOAT,
    DECIMAL_NN: GD_DECIMAL,
    DECIMAL_N: GD_DECIMAL,
    INTEGER_NN: GD_INTEGER,
    INTEGER_N: GD_INTEGER,
    SMALLINT_NN: GD_SMALLINT,
    SMALLINT_N: GD_SMALLINT,
    ARRAY_1D_NN: GD_DEFAULT,
    ARRAY_1D_N: GD_DEFAULT,
    ARRAY_ND_NN: GD_DEFAULT,
    ARRAY_ND_N: GD_DEFAULT,
    BIGINT_NN: GD_BIGINT,
    BIGINT_N: GD_BIGINT,
    VARBYTE_NN: GD_VARCHAR,
    VARBYTE_N: GD_VARCHAR,
    BYTE_NN: GD_CHAR,
    BYTE_N: GD_CHAR,
    LONG_VARBYTE_NN: GD_VARCHAR,
    LONG_VARBYTE_N: GD_VARCHAR,
    DATE_NNA: GD_DEFAULT,
    DATE_NA: GD_DEFAULT,
    DATE_NN: GD_DATE,
    DATE_N: GD_DATE,
    BYTEINT_NN: GD_BYTEINT,
    BYTEINT_N: GD_BYTEINT,
    TIME_NN: GD_CHAR,
    TIME_N: GD_CHAR,
    TIMESTAMP_NN: GD_CHAR,
    TIMESTAMP_N: GD_CHAR,
    TIME_NNZ: GD_CHAR,
    TIME_NZ: GD_CHAR,
    TIMESTAMP_NNZ: GD_CHAR,
    TIMESTAMP_NZ: GD_CHAR,
    INTERVAL_YEAR_NN: GD_DEFAULT,
    INTERVAL_YEAR_N: GD_DEFAULT,
    INTERVAL_YEAR_TO_MONTH_NN: GD_DEFAULT,
    INTERVAL_YEAR_TO_MONTH_N: GD_DEFAULT,
    INTERVAL_MONTH_NN: GD_DEFAULT,
    INTERVAL_MONTH_N: GD_DEFAULT,
    INTERVAL_DAY_NN: GD_DEFAULT,
    INTERVAL_DAY_N: GD_DEFAULT,
    INTERVAL_DAY_TO_HOUR_NN: GD_DEFAULT,
    INTERVAL_DAY_TO_HOUR_N: GD_DEFAULT,
    INTERVAL_DAY_TO_MINUTE_NN: GD_DEFAULT,
    INTERVAL_DAY_TO_MINUTE_N: GD_DEFAULT,
    INTERVAL_DAY_TO_SECOND_NN: GD_DEFAULT,
    INTERVAL_DAY_TO_SECOND_N: GD_DEFAULT,
    INTERVAL_HOUR_NN: GD_DEFAULT,
    INTERVAL_HOUR_N: GD_DEFAULT,
    INTERVAL_HOUR_TO_MINUTE_NN: GD_DEFAULT,
    INTERVAL_HOUR_TO_MINUTE_N: GD_DEFAULT,
    INTERVAL_HOUR_TO_SECOND_NN: GD_DEFAULT,
    INTERVAL_HOUR_TO_SECOND_N: GD_DEFAULT,
    INTERVAL_MINUTE_NN: GD_DEFAULT,
    INTERVAL_MINUTE_N: GD_DEFAULT,
    INTERVAL_MINUTE_TO_SECOND_NN: GD_DEFAULT,
    INTERVAL_MINUTE_TO_SECOND_N: GD_DEFAULT,
    INTERVAL_SECOND_NN: GD_DEFAULT,
    INTERVAL_SECOND_N: GD_DEFAULT,
    PERIOD_DATE_NN: GD_DEFAULT,
    PERIOD_DATE_N: GD_DEFAULT,
    PERIOD_TIME_NN: GD_DEFAULT,
    PERIOD_TIME_N: GD_DEFAULT,
    PERIOD_TIME_NNZ: GD_DEFAULT,
    PERIOD_TIME_NZ: GD_DEFAULT,
    PERIOD_TIMESTAMP_NN: GD_DEFAULT,
    PERIOD_TIMESTAMP_N: GD_DEFAULT,
    PERIOD_TIMESTAMP_NNZ: GD_DEFAULT,
    PERIOD_TIMESTAMP_NZ: GD_DEFAULT,
    XML_TEXT_NN: GD_DEFAULT,
    XML_TEXT_N: GD_DEFAULT,
    XML_TEXT_DEFERRED_NN: GD_DEFAULT,
    XML_TEXT_DEFERRED_N: GD_DEFAULT,
    XML_TEXT_LOCATOR_NN: GD_DEFAULT,
    XML_TEXT_LOCATOR_N: GD_DEFAULT,
}



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import datetime
import struct
try:
    import ujson as json
except ImportError:
    import json

from .constants import *
from .errors import *
from .fmt import *
from .logging import *
from .utils import *
from .types import *

from ._compat import *


__all__ = ['check_input', 'CharHandler', 'DateHandler', 'DecimalHandler', 'FloatHandler',
    'null_handler', 'dict_to_json', 'python_to_bytes', 'python_to_dict', 'python_to_strings',
    'strings_to_text', 'teradata_to_archive', 'text_to_strings', 'teradata_to_python',
    'python_to_sql', 'python_to_teradata', 'unpack', 'unpack_from', 'unpack_integer',
    'unpack_stmt_info']


FORMAT_LENGTH = {1: "b", 2: "h", 4: "i", 8 : "q", 16: "Qq"}


class Handler(object):
    handlers = []

    def __init__(self, columns):
        if not isinstance(columns, Columns):
            raise GiraffeEncodeError("Type '{}' must be type giraffez.types.Columns".format(type(columns)))
        self.columns = columns
        self._columns = []
        for index, column in enumerate(self.columns):
            for cond, fn in self.handlers:
                if column.type in cond:
                    self._columns.append((fn, index))

    def __call__(self, data):
        for fn, index in self._columns:
            if data[index] is not None:
                data[index] = fn(data[index])
        return data


class CharHandler(Handler):
    handlers = [
        (CHAR_TYPES | VAR_TYPES, replace_cr),
    ]


class DateHandler(Handler):
    handlers = [
        (DATE_TYPES, Date.from_integer),
        (TIME_TYPES, Time.from_string),
        (TIMESTAMP_TYPES, Timestamp.from_string)
    ]


class DecimalHandler(Handler):
    handlers = [
        (DECIMAL_TYPES, Decimal),
    ]


class FloatHandler(Handler):
    handlers = [
        (DECIMAL_TYPES, pipeline([float, str])),
    ]


def check_input(columns, items):
    if len(items) != len(columns):
        error_message = "Columns list contains {} columns and {} items provided.".format(len(columns), len(items))
        raise GiraffeEncodeError("{}: \n\t{!r}".format(error_message, items))

def null_handler(null=None):
    nulls = {"", None, null}
    def _handler(items):
        return [item if item not in nulls else None for item in items]
    return _handler

def dict_to_json(record):
    try:
        return json.dumps(record, ensure_ascii=False)
    except UnicodeDecodeError as error:
        log.debug("Debug[2]", "UnicodeDecodeError: {}".format(error.message))

def python_to_bytes(null):
    null = ensure_bytes(null)
    def _encoder(items):
        return [ensure_bytes(x) if x is not None else null for x in items]
    return _encoder

def python_to_dict(columns):
    def _encoder(items):
        return {x.name: y for x, y in zip(columns, items)}
    return _encoder

def python_to_strings(null):
    def _encoder(items):
        return [str(x) if x is not None else null for x in items]
    return _encoder

def strings_to_text(delimiter):
    def _encoder(items):
        return delimiter.join(items)
    return _encoder

def teradata_to_archive(data):
    return struct.pack("H", len(data)) + data

def text_to_strings(delimiter):
    if delimiter is None:
        _encoder = lambda s: [s.strip("\r\n")]
    else:
        _encoder = lambda s: s.strip("\r\n").split(delimiter)
    return _encoder

def teradata_to_python(columns):
    def _encoder(data):
        items = []
        indicator, data = data[0:columns.header_length], data[columns.header_length:]
        for column, nullable in zip(columns, Bytes.read(indicator)):
            if column.type in ALL_INTEGER_TYPES:
                value, data = unpack_integer(data[0:column.length], column.length), data[column.length:]
            elif column.type in FLOAT_TYPES:
                value, data = struct.unpack("d", data[0:column.length])[0], data[column.length:]
            elif column.type in DECIMAL_TYPES:
                if column.length > 8:
                    lower, upper = struct.unpack(FORMAT_LENGTH.get(column.length), data[0:column.length])
                    value = ((upper << 64) | lower)
                else:
                    value = struct.unpack(FORMAT_LENGTH.get(column.length), data[0:column.length])[0]
                data = data[column.length:]
                if column.scale > 0:
                    (a, b) = divmod(abs(value), 10 ** column.scale)
                    if value < 0:
                        a = -a
                    value = "{}.{:0{}}".format(a, b, column.scale)
                else:
                    value = "{}".format(value)
            elif column.type in CHAR_TYPES:
                value, data = data[0:column.length], data[column.length:]
                value = value.strip().replace("\r", "\n")
            elif column.type in VAR_TYPES:
                length = unpack_integer(data[0:2], 2)
                value, data = data[2:2+length], data[2+length:]
                value = value.replace("\r", "\n")
            elif column.type in DATE_TYPES:
                n, data = unpack_integer(data[0:column.length], column.length), data[column.length:]
                value = Date.from_integer(n)
            elif column.type in TIME_TYPES:
                value, data = Time.from_string(data[0:column.length]), data[column.length:]
            elif column.type in TIMESTAMP_TYPES:
                value, data = Timestamp.from_string(data[0:column.length]), data[column.length:]
            else:
                value, data = data[0:column.length], data[column.length:]
            if nullable:
                value = None
            items.append(value)
        return items
    return _encoder

def python_to_sql(table_name, columns, date_conversion=True):
    if date_conversion:
        def convert_date(s):
            value = Date.from_string(s)
            if value is None:
                raise GiraffeEncodeError("Cannot convert date '{}'".format(s))
            return value.to_string()

        def convert_datetime(s, length):
            if isinstance(s, datetime.date):
                value = Timestamp.from_datetime(s)
            else:
                value = Timestamp.from_string(s)
            if value is None:
                raise GiraffeEncodeError("Cannot convert datetime '{}'".format(s))
            return value.to_string(length)
    else:
        convert_date = lambda a: a
        convert_datetime = lambda a, b: a
    def _encoder(items):
        check_input(columns, items)
        values = []
        for item, column in zip(items, columns):
            if item is None:
                continue
            if column.type in CHAR_TYPES | VAR_TYPES:
                value = quote_string(escape_quotes(item))
            elif column.type in DATE_TYPES:
                if isinstance(item, datetime.date):
                    value = str(Date.from_datetime(item).to_integer())
                else:
                    value = quote_string(convert_date(escape_quotes(item)))
            elif column.type in TIME_TYPES | TIMESTAMP_TYPES:
                value = quote_string(convert_datetime(item, column.length))
            else:
                value = str(item)
            values.append((column.name, value))
        return "ins into {} ({}) values ({});".format(table_name,
            ",".join(quote_string(x[0], '"') for x in values),
            ",".join(x[1] for x in values))
    return _encoder

def python_to_teradata(columns, allow_precision_loss=False):
    if allow_precision_loss:
        _float_handler = FloatHandler(columns)
    else:
        _float_handler = lambda a: a
    def _encoder(items):
        check_input(columns, items)
        items = _float_handler(items)
        indicator = Bytes(columns)
        data = [indicator]
        for i, (item, column) in enumerate(zip(items, columns)):
            if item is None:
                indicator.set_bit(i)
                if column.type in STRING_NULLS:
                    value = b"\x20" * column.length
                elif column.type in VARIABLE_NULLS:
                    value = b"\x00\x00"
                else:
                    value = b"\x00" * column.length
                data.append(value)
                continue
            if column.type in ALL_INTEGER_TYPES:
                pack_char = FORMAT_LENGTH.get(column.length)
                data.append(struct.pack(pack_char, int(item)))
            elif column.type in FLOAT_TYPES:
                data.append(struct.pack("d", float(item)))
            elif column.type in DECIMAL_TYPES:
                try:
                    parts = item.split(".", 1)
                except AttributeError as error:
                    raise GiraffeEncodeError("Invalid value for Decimal data type: {}".format(item))
                if len(parts) == 1 or column.scale == 0:
                     a = parts[0]
                     b = ""
                elif len(parts) == 2:
                    a = parts[0]
                    b = parts[1].replace(".", "")
                try:
                    item = int("{}{}".format(a, b.ljust(column.scale, "0")))
                except ValueError as error:
                    raise GiraffeEncodeError("Invalid value for Decimal data type: {}".format(item))
                pack_char = FORMAT_LENGTH.get(column.length)
                if column.length > 8:
                    upper = item >> 64
                    lower = (upper << 64) ^ item
                    value = struct.pack(pack_char, lower, upper)
                else:
                    value = struct.pack(pack_char, item)
                data.append(value)
            elif column.type in CHAR_TYPES:
                item = ensure_bytes(item[:column.length])
                value = item.ljust(column.length, b"\x20")
                data.append(value)
            elif column.type in VAR_TYPES:
                data.append(struct.pack("H", len(item)) + ensure_bytes(item))
            elif column.type in DATE_TYPES:
                date_value = Date.from_string(item)
                if date_value:
                    value = struct.pack("i", int(date_value))
                else:
                    value = None
                if value is None:
                    indicator.set_bit(i)
                    value = b"\x20" * column.length
                data.append(value)
            else:
                data.append(item)
        return b"".join(ensure_bytes(x) for x in data)
    return _encoder

def unpack(fmt, data):
    """
    Unpack binary data based on struct format string. Simulates
    reading from a file-like object by slicing the data string that is
    returned.

    :return: At tuple of (unpacked value, remaining data)
    :rtype: tuple
    """
    length = struct.calcsize(fmt)
    value = struct.unpack(fmt, data[:length])[0]
    data = data[length:]
    return value, data

def unpack_from(fmt, data):
    """
    Unpack binary data based on struct format string. Simulates
    reading from a file-like object by slicing the data string that is
    returned. Differs from :func:`unpack` because it assumes a packed
    value represents the length of a variable-length value that
    follows.

    :return: At tuple of (unpacked value, remaining data)
    :rtype: tuple
    """
    length = unpack(fmt, data)[0]
    data = data[struct.calcsize(fmt):]
    value = data[:length]
    data = data[length:]
    return value, data

def unpack_integer(s, length):
    return struct.unpack(FORMAT_LENGTH.get(length), s)[0]

def unpack_stmt_info(data):
    columns = []
    while len(data) > 0:
        ext_layout, ext_type, ext_len = struct.unpack("hhh", data[:6])
        data = data[6:]
        ext_data = data[:ext_len]
        data = data[ext_len:]
        if ext_layout != 1:
            continue
        db, ext_data = unpack_from("h", ext_data)
        table, ext_data = unpack_from("h", ext_data)
        column_name, ext_data = unpack_from("h", ext_data)
        some_thang, ext_data = unpack("h", ext_data)
        useless, ext_data = unpack_from("h", ext_data)
        column_title, ext_data = unpack_from("h", ext_data)
        column_format, ext_data = unpack_from("h", ext_data)
        default_value, ext_data = unpack_from("h", ext_data)
        identity_column, ext_data = unpack("s", ext_data)
        mosdef_writable, ext_data = unpack("s", ext_data)
        not_not_null, ext_data = unpack("s", ext_data)
        can_ret_null, ext_data = unpack("s", ext_data)
        permitted_where, ext_data = unpack("s", ext_data)
        modifiable_column, ext_data = unpack("s", ext_data)
        column_type, ext_data = unpack("h", ext_data)
        user_def_type, ext_data = unpack("h", ext_data)
        type_name, ext_data = unpack_from("h", ext_data)
        data_type_misc_info, ext_data = unpack_from("h", ext_data)
        column_length, ext_data = unpack("Q", ext_data) #total_bytes_might_ret
        precision, ext_data = unpack("h", ext_data) #total_digits
        interval_digits, ext_data = unpack("h", ext_data) # if termporal data type
        scale, ext_data = unpack("h", ext_data) #fractional_digits
        char_set_type, ext_data = unpack("b", ext_data)
        total_bytes_ret, ext_data = unpack("Q", ext_data)
        case_sensitive, ext_data = unpack("s", ext_data)
        numeric_signed, ext_data = unpack("s", ext_data)
        uniquely_desc_row, ext_data = unpack("s", ext_data)
        column_only_member_uniq_index, ext_data = unpack("s", ext_data)
        is_expression, ext_data = unpack("s", ext_data)
        permitted_orderby, ext_data = unpack("s", ext_data)
        if column_type in [DECIMAL_N, DECIMAL_NN]:
            if precision <= 2:
                column_length = 1
            elif precision <= 4:
                column_length = 2
            elif precision <= 9:
                column_length = 4
            elif precision <= 18:
                column_length = 8
            else:
                column_length = 16
        else:
            precision = 0
            scale = 0
        if column_name == "":
            column_name = column_title
        columns.append({
            "name": column_name,
            "type": column_type,
            "length": column_length,
            "precision": precision,
            "scale": scale,
            "nullable": can_ret_null,
            "default": default_value,
            "format": column_format
        })
    return columns



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
import hashlib
import base64

from Crypto.Cipher import AES

from ._compat import *


__all__ = ['create_key_file', 'generate_key', 'Crypto']


def create_key_file(path):
    """
    Creates a new encryption key in the path provided and sets the file
    permissions.  Setting the file permissions currently does not work
    on Windows platforms because of the differences in how file
    permissions are read and modified.
    """
    iv = "{}{}".format(os.urandom(32), time.time())
    new_key = generate_key(ensure_bytes(iv))
    with open(path, "wb") as f:
        f.write(base64.b64encode(new_key))
    os.chmod(path, 0o400)

def generate_key(s):
    for i in range(65536):
        s = hashlib.sha256(s).digest()
    return s


class Crypto(object):
    def __init__(self, key, bs=16): 
        self.key = hashlib.sha256(key).digest()
        self.bs = bs

    def decrypt(self, enc):
        enc = base64.b64decode(enc)
        iv = enc[:self.bs]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self.unpad(cipher.decrypt(enc[self.bs:])).decode("utf-8")

    def encrypt(self, raw):
        raw = self.pad(raw)
        iv = os.urandom(self.bs)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(raw))

    def pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    def unpad(self, s):
        return s[:-ord(s[len(s)-1:])]

    @classmethod
    def from_key_file(cls, path):
        with open(path, "rb") as f:
            return cls(base64.b64decode(f.read()))



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class GiraffeError(Exception):
    """
    Baseclass for all giraffez errors.
    """

class GeneralError(GiraffeError):
    """
    General giraffez error.
    """

class GiraffeNotFound(GiraffeError):
    """
    Raised when giraffez C modules not found.
    """

class TeradataCLIv2NotFound(GiraffeNotFound):
    """
    Unable to import giraffez._cli.  This indicates that either the
    giraffez C extensions did not compile correctly, or more likely,
    there is an issue with the environment or installation of the
    Teradata dependencies.  The Teradata Call-Level Interface Version 2
    requires several environment variables to be set to find the shared
    library files and error message catalog.

    For more information, refer to this section in the giraffez
    documentation:
        http://www.capitalone.io/giraffez/intro.html#environment.
    """

class TeradataPTAPINotFound(GiraffeNotFound):
    """
    Unable to import giraffez._tpt.  This indicates that either the
    giraffez C extensions did not compile correctly, or more likely,
    there is an issue with the environment or installation of the
    Teradata dependencies.  The Teradata Parallel Transporter API
    requires several environment variables to be set to find the shared
    library files and error message catalog.

    For more information, refer to this section in the giraffez
    documentation:
        http://www.capitalone.io/giraffez/intro.html#environment.
    """

class GiraffeTypeError(GiraffeError):
    """
    Baseclass for all giraffez type errors.
    """

class GiraffeEncodeError(GiraffeError):
    """
    Raised when unable to encode the provided object.
    """

class ConfigurationError(GiraffeError): 
    """
    For use with configuration file handling. 
    """

class FileNotFound(GiraffeError): 
    """
    Raised when file does not exist.
    """

class ConfigNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class KeyNotFound(ConfigurationError, FileNotFound):
    """
    Raised when the specified configuration file does not exist.
    """

class ConfigReadOnly(ConfigurationError):
    """
    Raised when a write is attempted on a configuration file was opened
    in read mode.
    """

class ConnectionLock(ConfigurationError):
    """
    Raised when connection is locked by invalid attempts and the
    'protect' feature is being used.
    """

    def __init__(self, dsn):
        super(ConnectionLock, self).__init__(("Connection {0} is currently locked. please update "
            "credentials and run:\n\tgiraffez config --unlock {0}").format(dsn))

class TeradataError(Exception): 
    """
    Baseclass for all Teradata errors. This exception represents any
    error that originates from Teradata CLIv2/TPTAPI.

    This will attempt to parse error codes from the message in the case
    that the error is a Teradata CLIv2 error.
    """

    def __init__(self, message, code=None):
        super(TeradataError, self).__init__(message)

        if isinstance(message, Exception):
            message = getattr(message, 'message', '')

        if code is None:
            if isinstance(message, (list, tuple)):
                code, message = message[0], message[1]
            else:
                if ":" in message:
                    parts = [x.strip() for x in message.split(":", 1)]
                    try:
                        code = int(parts[0])
                        message = parts[1]
                    except ValueError as error:
                        code = None
                        message = "Unable to parse CLI error code/message"
                else:
                    code = None
                    message = "Unable to parse CLI error code/message"
            if code == ObjectDoesNotExist.code:
                raise ObjectDoesNotExist(message, code)
            elif code == ObjectNotTable.code:
                raise ObjectNotTable(message, code)
            elif code == ObjectNotView.code:
                raise ObjectNotView(message, code)
            elif code == TransactionAborted.code:
                raise TransactionAborted(message, code)
            elif code == CannotReleaseMultiLoad.code:
                raise CannotReleaseMultiLoad(message, code)
            elif code == MultiLoadTableExists.code:
                raise MultiLoadTableExists(message, code)
            elif code == MultiLoadWorkTableNotFound.code:
                raise MultiLoadWorkTableNotFound(message, code)
            elif code == InvalidCredentialsError.code:
                raise InvalidCredentialsError(message, code)

        self.message = message
        self.code = code

    def __repr__(self):
        return "{}: {}".format(self.code, self.message)

class ObjectDoesNotExist(TeradataError):
    """
    Teradata object does not exist.
    """
    code = 3807

class ObjectNotTable(TeradataError):
    """
    Teradata object not a table.
    """
    code = 3853

class ObjectNotView(TeradataError):
    """
    Teradata object not a view.
    """
    code = 3854

class TransactionAborted(TeradataError):
    """
    Teradata transaction aborted.
    """
    code = 2631

class MultiLoadError(TeradataError):
    """
    General MultiLoad error class.
    """

class CannotReleaseMultiLoad(MultiLoadError):
    """
    Raised when MultiLoad cannot be released because it is in the
    apply phase.
    """
    code = 2572

class MultiLoadTableExists(MultiLoadError):
    """
    Raised when MultiLoad worktables exist.
    """
    code = 2574

class MultiLoadLocked(MultiLoadError):
    """
    Raised when MultiLoad appears to be locked. Used to determine when
    MultiLoad should be released.
    """

class MultiLoadWorkTableNotFound(MultiLoadLocked):
    """
    Raised when MultiLoad worktable is missing during restart.
    """
    code = 2583

class InvalidCredentialsError(TeradataError):
    """
    Raised when connection credentials are incorrect.
    """
    code = 8017



import os
import sys
import io
import logging
import re
import platform
import traceback
import multiprocessing
from setuptools import setup, Extension as _Extension, find_packages
from distutils.command.build_ext import build_ext
from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError


PY3 = sys.version_info[0] == 3

if not PY3:
    FileNotFoundError = (IOError, OSError)


class GiraffeBuildError(Exception):
    """
    Baseclass for all build errors.
    """

class PlatformNotSupported(GiraffeBuildError):
    """
    Raised when the installation platform is unsupported or unknown.
    """

class TeradataNotFound(GiraffeBuildError):
    """
    Raised when unable to automatically find either Teradata CLIv2 or
    Teradata PT API files.
    """


def fix_compile(remove_flags):
    """
    Monkey-patch compiler to allow for removal of default compiler flags.
    """
    import distutils.ccompiler

    def _fix_compile(self, sources, output_dir=None, macros=None, include_dirs=None, debug=0,
            extra_preargs=None, extra_postargs=None, depends=None):
        for flag in remove_flags:
            if flag in self.compiler_so:
                self.compiler_so.remove(flag)
        macros, objects, extra_postargs, pp_opts, build = self._setup_compile(output_dir, macros,
                include_dirs, sources, depends, extra_postargs)
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
        for obj in objects:
            try:
                src, ext = build[obj]
            except KeyError:
                continue
            self._compile(obj, src, ext, cc_args, extra_postargs, pp_opts)
        return objects

    distutils.ccompiler.CCompiler.compile = _fix_compile

# strict-prototypes is only valid for C code so causes invalid warnings
# with C++.
# unreachable-code is considered to be a broken warning and has been
# removed from newer versions of gcc.
remove_flags = ['-Wstrict-prototypes', '-Wunreachable-code']
fix_compile(remove_flags)

def is_64bit():
    if sys.maxsize > 2**32:
        return True
    return False

def find_teradata_home():
    """
    Attempts to find the Teradata install directory with the defaults
    for a given platform.  Should always return `None` when the defaults
    are not present and the TERADATA_HOME environment variable wasn't
    explicitly set to the correct install location.
    """
    if platform.system() == 'Windows':
        # The default installation path for Windows is split between the
        # Windows directories for 32-bit/64-bit applications.  It is
        # worth noting that Teradata archiecture installed should match
        # the architecture of the Python architecture being used (i.e.
        # TTU 32-bit is required /w Python 32-bit and TTU 64-bit is
        # required for Python 64-bit).
        if is_64bit():
            return latest_teradata_version("C:/Program Files/Teradata/Client")
        else:
            return latest_teradata_version("C:/Program Files (x86)/Teradata/Client")
    elif platform.system() == 'Linux':
        return latest_teradata_version("/opt/teradata/client")
    elif platform.system() == 'Darwin':
        return latest_teradata_version("/Library/Application Support/teradata/client")
    else:
        # In the case nothing is found, the default for Linux is
        # attempted as a last effort to find the correct install
        # directory.
        return latest_teradata_version("/opt/teradata/client")

def latest_teradata_version(search_directory):
    """
    Attempts to find the latest version of Teradata installed based upon
    the default Teradata installation directory structure.
    """
    try:
        directories = []
        for d in os.listdir(search_directory):
            if re.match("[\d\.]+", d):
                directories.append(d)
        if not directories:
            return None
        return os.path.join(search_directory, max(directories))
    except FileNotFoundError:
        sys.stderr.write("Teradata FileNotFound.")
        return None

def giraffez_version():
    version_regex = re.compile(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', re.MULTILINE)

    with open('giraffez/__init__.py', 'r') as f:
        return version_regex.search(f.read()).group(1)


TERADATA_HOME = os.environ.get('TERADATA_HOME', find_teradata_home())


class Extension(_Extension):
    name = None

    objects = []

    depends_on = []

    compiled = False
    success = False
    shared_object = True

    def __init__(self, *args, **kwargs):
        _Extension.__init__(self, self.name, self.sources, **kwargs)

        # Add local source files to include_dirs
        self.include_dirs.append(os.path.join(os.getcwd(), "giraffez"))

        # Windows compatbility
        if platform.system() == 'Windows':
            if is_64bit():
                self.define_macros.append(('WIN64', 1))
            else:
                self.define_macros.append(('WIN32', 1))
        else:
            self.extra_compile_args = ['-Wfatal-errors']

    @classmethod
    def set_objects(cls, objects):
        cls.objects = objects

    @property
    def language(self):
        if any(f.endswith(".cc") for f in self.sources):
            return "c++"
        return "c"

    @language.setter
    def language(self, value):
        pass

    def setup(self):
        pass


class CommonExtension(Extension):
    name = "giraffez._common"
    sources = [
        "giraffez/commonmodule.c",
    ]


class EncoderExtension(Extension):
    name = "giraffez._encoder"

    sources = [
        "giraffez/encoder/convert.c",
        "giraffez/encoder/types.c",
        "giraffez/encoder/unpack.c",
        "giraffez/encoder/encoderobject.c",
        "giraffez/encodermodule.c",
    ]

    depends = [
        "giraffez/encoder/convert.h",
        "giraffez/encoder/types.h",
        "giraffez/encoder/unpack.h",
        "giraffez/encoder/encoderobject.h",
        "giraffez/encodermodule.c",
    ]


class CLIExtension(Extension):
    name = "giraffez._cli"

    sources = [
        "giraffez/cli/cmdobject.c",
        "giraffez/climodule.c"
    ]

    depends = [
        "giraffez/cli/cmdobject.h",
        "giraffez/climodule.c"
    ]

    cli_include_dir = None
    cli_library_dir = None

    def setup(self):
        if TERADATA_HOME is None:
            raise TeradataNotFound("Unable to find the Teradata files.")

        if platform.system() == 'Windows':
            if is_64bit():
                tdcli_inc = os.path.join(TERADATA_HOME, "CLIv2/inc")
                tdcli_lib = os.path.join(TERADATA_HOME, "CLIv2/lib")
            else:
                tdcli_inc = os.path.join(TERADATA_HOME, "CLIv2/inc")
                tdcli_lib = os.path.join(TERADATA_HOME, "CLIv2/lib")
        elif platform.system() == 'Linux':
            if is_64bit():
                tdcli_inc = os.path.join(TERADATA_HOME, "include")
                tdcli_lib = os.path.join(TERADATA_HOME, "lib64")
            else:
                tdcli_inc = os.path.join(TERADATA_HOME, "include")
                tdcli_lib = os.path.join(TERADATA_HOME, "lib")
        elif platform.system() == 'Darwin':
            if is_64bit():
                tdcli_inc = os.path.join(TERADATA_HOME, "include")
                tdcli_lib = os.path.join(TERADATA_HOME, "lib")
            else:
                tdcli_inc = os.path.join(TERADATA_HOME, "include")
                tdcli_lib = os.path.join(TERADATA_HOME, "lib")
        else:
            raise PlatformNotSupported("This following platform is unsupported or unknown: '{}'".format(platform.system()))

        if os.path.isdir(tdcli_inc):
            self.cli_include_dir = tdcli_inc
        if os.path.isdir(tdcli_lib):
            self.cli_library_dir = tdcli_lib

        if not self.cli_include_dir or not self.cli_library_dir:
            raise TeradataNotFound("Cannot find the Teradata CLIv2 files.")

        self.include_dirs.append(self.cli_include_dir)
        self.library_dirs.append(self.cli_library_dir)

        # Set libraries
        if platform.system() == 'Windows':
            self.libraries.append("wincli32")
        elif platform.system() == 'Linux':
            self.libraries.append("cliv2")
        elif platform.system() == 'Darwin':
            self.libraries.append("cliv2")


class TPTExtension(Extension):
    name = "giraffez._tpt"

    sources = [
        "giraffez/tpt/exportobject.cc",
        "giraffez/tpt/loadobject.cc",
        "giraffez/tptmodule.cc"
    ]

    depends = [
        "giraffez/tpt/exportobject.h",
        "giraffez/tpt/loadobject.h",
        "giraffez/tptmodule.cc"
    ]

    depends_on = [EncoderExtension]

    tpt_include_dir = None
    tpt_library_dir = None

    def setup(self):
        if TERADATA_HOME is None:
            raise TeradataNotFound("Unable to find the Teradata files.")

        if platform.system() == 'Windows':
            if is_64bit():
                tptapi_inc = os.path.join(TERADATA_HOME, "Teradata Parallel Transporter/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "Teradata Parallel Transporter/bin64")
            else:
                tptapi_inc = os.path.join(TERADATA_HOME, "Teradata Parallel Transporter/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "Teradata Parallel Transporter/bin")
            if not os.path.isdir(tptapi_lib):
                tptapi_lib = os.path.join(TERADATA_HOME, "bin")
        elif platform.system() == 'Linux':
            if is_64bit():
                tptapi_inc = os.path.join(TERADATA_HOME, "tbuild/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "tbuild/lib64")
            else:
                tptapi_inc = os.path.join(TERADATA_HOME, "tbuild/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "tbuild/lib")
        elif platform.system() == 'Darwin':
            if is_64bit():
                tptapi_inc = os.path.join(TERADATA_HOME, "tbuild/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "tbuild/lib")
            else:
                tptapi_inc = os.path.join(TERADATA_HOME, "tbuild/tptapi/inc")
                tptapi_lib = os.path.join(TERADATA_HOME, "tbuild/lib")
        else:
            raise PlatformNotSupported("This following platform is unsupported or unknown: '{}'".format(platform.system()))

        if os.path.isdir(tptapi_inc):
            self.tpt_include_dir = tptapi_inc
        if os.path.isdir(tptapi_lib):
            self.tpt_library_dir = tptapi_lib

        if not self.tpt_include_dir or not self.tpt_library_dir:
            raise TeradataNotFound("Cannot find the Teradata Parallel Transporter API files.")

        self.include_dirs.append(self.tpt_include_dir)
        self.library_dirs.append(self.tpt_library_dir)

        if platform.system() == 'Windows':
            self.libraries.append("telapi")
        elif platform.system() == 'Linux':
            self.libraries.append("telapi")
            self.libraries.append("pxicu")
        elif platform.system() == 'Darwin':
            self.libraries.append("telapi")
            self.libraries.append("pxicu")


class BuildExt(build_ext):
    cache = {}

    def run(self):
        # Disabling parallel build for now. It causes issues on multiple
        # platforms with concurrent file access causing odd build errors
        #self.parallel = multiprocessing.cpu_count()
        build_ext.run(self)

    def get_inplace_path(self, ext_name):
        fullname = self.get_ext_fullname(ext_name)
        modpath = fullname.split('.')
        filename = self.get_ext_filename(modpath[-1])

        # the inplace option requires to find the package directory
        # using the build_py command for that
        package = '.'.join(modpath[0:-1])
        build_py = self.get_finalized_command('build_py')
        package_dir = os.path.abspath(build_py.get_package_dir(package))

        # returning
        #   package_dir/filename
        fullpath = os.path.join(package_dir, filename)
        return fullpath

    def compile(self, ext):
        objects = self.compiler.compile(ext.sources,
            output_dir=os.path.join(os.getcwd(), "build"),
            macros=ext.define_macros,
            include_dirs=ext.include_dirs,
            extra_postargs=ext.extra_compile_args,
            depends=ext.depends)
        self.cache[ext.name] = objects
        return objects

    def build_extension(self, ext):
        ext.setup()
        objects = self.compile(ext)
        ext_path = self.get_ext_fullpath(ext.name)
        if ext.depends_on:
            for dep in ext.depends_on:
                if dep.name not in self.cache:
                    objects += self.compile(dep())
                else:
                    objects += self.cache[dep.name]

        # Return early when a shared object is not being created. This
        # is useful when dealing with Extensions that need to have
        # object files built for another Extension that will be creating
        # a shared object.
        if not ext.shared_object:
            return

        self.compiler.link_shared_object(
            objects,
            ext_path,
            libraries=self.get_libraries(ext),
            library_dirs=ext.library_dirs,
            runtime_library_dirs=ext.runtime_library_dirs,
            extra_postargs=ext.extra_link_args,
            export_symbols=self.get_export_symbols(ext),
            debug=self.debug,
            build_temp=self.build_temp,
            target_lang=ext.language)

        # This ensures that the shared objects are *also* built
        # in the path to avoid problems with path resolution when
        # working with local files.
        src = self.get_ext_fullpath(ext.name)
        dst = self.get_inplace_path(ext.name)
        self.mkpath(os.path.dirname(dst))
        self.copy_file(src, dst, preserve_mode=False)


if __name__ == '__main__':
    ext_modules = [CommonExtension(), EncoderExtension(), CLIExtension(), TPTExtension()]

    with open('requirements.txt') as f:
        requirements = f.read().splitlines()

    with io.open("README.rst", encoding="utf-8") as f:
        long_description = f.read()
        long_description = long_description.replace('logo.png', 'logo-md.png')

    setup(
        name="giraffez",
        description="a user-friendly and fast Teradata client for Python",
        long_description=long_description,
        license="Apache 2.0",
        url="https://github.com/capitalone/giraffez",
        version=giraffez_version(),
        packages=find_packages(exclude=['tests*']),
        ext_modules=ext_modules,
        entry_points={
            'console_scripts': [
                'giraffez = giraffez.__main__:main'
            ]
        },
        zip_safe=False,
        package_dir={'giraffez': 'giraffez'},
        cmdclass={'build_ext': BuildExt},
        install_requires=requirements,
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Programming Language :: C',
            'Programming Language :: C++',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.1',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: POSIX :: Linux',
            'Topic :: Database',
            'Topic :: Software Development',
            'Topic :: Utilities'
        ]
    )



# -*- coding: utf-8 -*-
#
# Copyright 2016 Capital One Services, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

TPT_NOT_FOUND = False
try:
    from . import _tpt
except ImportError:
    TPT_NOT_FOUND = True

from collections import defaultdict

from .config import *
from .connection import *
from .constants import *
from .encoders import *
from .errors import *
from .fmt import *
from .io import *
from .logging import *
from .parser import *
from .sql import *
from .types import *
from .utils import *

from ._compat import *


__all__ = ['TeradataExport']


class TeradataExport(Connection):
    """
    The class for using the Teradata Parallel Transport API to quickly export
    large amounts of data.

    Exposed under the alias :class:`giraffez.Export`.

    :param str query: Either SQL query to execute and return results of, **or** the
        name of a table to export in its entirety
    :param str host: Omit to read from :code:`~/.girafferc` configuration file.
    :param str username: Omit to read from :code:`~/.girafferc` configuration file.
    :param str password: Omit to read from :code:`~/.girafferc` configuration file.
    :param str delimiter: The delimiter to use when exporting with the 'text'
        encoding. Defaults to '|'
    :param str null: The string to use to represent a null value with the
        'text' encoding. Defaults to 'NULL'
    :param str encoding: The encoding format in which to export the data.
        Defaults to 'text'.  Possible values are 'text' - delimited text 
        output, 'dict' - to output each row as a :code:`dict` object mapping column names
        to values, 'json' - to output each row as a JSON encoded string, and 
        'archive' to output in giraffez archive format
    :param int log_level: Specify the desired level of output from the job.
        Possible values are :code:`giraffez.SILENCE` :code:`giraffez.INFO` (default),
        :code:`giraffez.VERBOSE` and :code:`giraffez.DEBUG`
    :param str config: Specify an alternate configuration file to be read from, when previous paramters are omitted.
    :param str key_file: Specify an alternate key file to use for configuration decryption
    :param string dsn: Specify a connection name from the configuration file to be
        used, in place of the default.
    :param bool protect: If authentication with Teradata fails and :code:`protect` is :code:`True` 
        locks the connection used in the configuration file. This can be unlocked using the
        command :code:`giraffez config --unlock <connection>` changing the connection password,
        or via the :meth:`~giraffez.config.Config.unlock_connection` method.
    :raises `giraffez.errors.InvalidCredentialsError`: if the supplied credentials are incorrect
    :raises `giraffez.errors.TeradataError`: if the connection cannot be established

    Meant to be used, where possible, with Python's :code:`with` context handler
    to guarantee that connections will be closed gracefully when operation
    is complete:

    .. code-block:: python

       with giraffez.Export('dbc.dbcinfo') as export:
           print(export.header)
           for row in export.results():
               print(row)
    """

    def __init__(self, query=None, host=None, username=None, password=None,
            delimiter=DEFAULT_DELIMITER, null=DEFAULT_NULL, encoding=DEFAULT_ENCODING,
            log_level=INFO, config=None, key_file=None, dsn=None, protect=False):
        if TPT_NOT_FOUND:
            raise TeradataPTAPINotFound(TeradataPTAPINotFound.__doc__.rstrip())
        super(TeradataExport, self).__init__(host, username, password, log_level, config, key_file,
            dsn, protect)

        #: Attributes used with property getter/setters
        self._columns = None
        self._query = None
        self._encoding = None

        self.delimiter = delimiter
        self.null = null
        self.encoding = encoding
        self.initiated = False

        #: Attributes set via property getter/setters
        self.query = query

    def _connect(self, host, username, password):
        self.export = _tpt.Export(host, username, password)
        self._handle_error()

    def _get_columns(self):
        if not self.initiated:
            self._initiate()
        columns = Columns(self.export.columns())
        _columns = defaultdict(int)
        for column in columns:
            if column.original_name in _columns:
                column.title = "{}_{}".format(column.original_name, _columns[column.original_name])
            _columns[column.original_name] += 1
        for column in columns:
            log.verbose("Debug[1]", repr(column))
        self.export.set_columns(columns)
        return columns

    def _handle_error(self):
        if self.export.status() >= TD_ERROR:
            if self.export.status() == CLI_ERR_INVALID_USER:
                if self.protect:
                    Config.lock_connection(self.config, self.dsn)
                raise InvalidCredentialsError(self.export.error_message())
            raise TeradataError(self.export.error_message())

    def _initiate(self):
        title, version = get_version_info()
        query_band = "UTILITYNAME={};VERSION={};".format(title, version)
        self.export.add_attribute(TD_QUERY_BAND_SESS_INFO, query_band)
        log.info("Export", "Initiating Teradata PT request (awaiting server)  ...")
        self.export.initiate()
        self._handle_error()
        log.info("Export", "Teradata PT request accepted.")
        self.export.get_schema()
        self._handle_error()
        self.initiated = True
        if self.encoding == "text":
            self.processor = identity
        elif self.encoding == "json":
            self.processor = dict_to_json
        elif self.encoding == "dict":
            self.processor = identity
        elif self.encoding == "archive":
            self.processor = identity

        self.options("encoding", self.encoding, 1)
        if self.encoding == "text":
            self.options("delimiter", self.delimiter, 2)
            self.options("null", self.null, 3)
        log.info("Export", "Executing ...")
        log.info(self.options)

    def close(self):
        log.info("Export", "Closing Teradata PT connection ...")
        self.export.close()
        log.info("Export", "Teradata PT request complete.")

    @property
    def columns(self):
        """
        Retrieve columns if necessary, and return them.

        :rtype: :class:`~giraffez.types.Columns`
        """
        if self._columns is None:
            self._columns = self._get_columns()
        return self._columns

    @property
    def encoding(self):
        return self._encoding

    @encoding.setter
    def encoding(self, value):
        if value == self._encoding:
            return
        else:
            self._encoding = value
        self.initiated = False
    
    @property
    def header(self):
        """
        Return the header for the resulting data.

        :return: String containing the formatted header, either column names
            separated by the current :code:`delimiter` value if :code:`encoding` is 'text',
            or the serialized :class:`~giraffez.types.Columns` object for
            'archive' encoding
        :rtype: str
        :raises `giraffez.errors.GiraffeError`: if the query or table
            has not been specified
        """
        if self.query is None:
            raise GiraffeError("Must set target table or query.")
        if self.encoding == "archive":
            return self.columns.serialize()
        return self.delimiter.join(self.columns.names)

    @property
    def query(self):
        """
        :return: The current query if it has been set, otherwise :code:`None`
        :rtype: str
        """
        return self._query

    @query.setter
    def query(self, query):
        """
        Set the query to be run and initiate the connection with Teradata.
        Only necessary if the query/table name was not specified as an argument
        to the constructor of the instance.

        :param str query: Valid SQL query to be executed
        """
        if query is None:
            return
        if log.level >= VERBOSE:
            self.options("query", query, 5)
        else:
            self.options("query", truncate(query), 5)
        statements = parse_statement(remove_curly_quotes(query))
        if not statements:
            raise GiraffeError("Unable to parse SQL statement")
        if len(statements) > 1:
            show_warning(("MORE THAN ONE STATEMENT RECEIVED, EXPORT OPERATIONS ALLOW ONE "
                "STATEMENT - ONLY THE FIRST STATEMENT WILL BE USED."), RuntimeWarning)
        statement = statements[0]
        log.debug("Debug[2]", "Statement (sanitized): {!r}".format(statement))
        if not (statement.startswith("select ") or statement.startswith("sel ")):
            statement = "select * from {}".format(statement)
        if statement == self.query:
            return
        else:
            self._query = statement
        self.initiated = False
        self.export.add_attribute(TD_SELECT_STMT, statement)
        # When new query is set refresh columns
        self._columns = None

    def results(self):
        """
        Generates and yields row results as they are returned from
        Teradata and processed. 
        
        :return: A generator that can be iterated over or coerced into a list.
        :rtype: iterator (yields ``string`` or ``dict``)

        .. code-block:: python

           # iterate over results
           with giraffez.Export('dbc.dbcinfo') as export:
               for row in export.results():
                   print(row)

           # retrieve results as a list
           with giraffez.Export('db.table2') as export:
               results = list(export.results())
        """
        if self.query is None:
            raise GiraffeError("Must set target table or query.")
        if not self.initiated:
            self._initiate()
            self._columns = self._get_columns()
        if self.encoding == "archive":
            def export_get_buffer():
                return self.export.get_buffer_raw()
        elif self.encoding in {"dict", "json"}:
            def export_get_buffer():
                return self.export.get_buffer_dict()
        else:
            def export_get_buffer():
                return self.export.get_buffer_str(self.null, self.delimiter)

        while True:
            data = export_get_buffer()
            if not data:
                break
            for row in data:
                yield self.processor(row)
